# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os,sys
from ferenda.compat import unittest

import codecs
from ferenda.textreader import TextReader

PREFIX = os.path.dirname(__file__)+"/files/textreader"
    
class Basic(unittest.TestCase):
    def setUp(self):
        self.f = TextReader(PREFIX + "/LICENSE.txt",linesep=TextReader.UNIX)

    def testReadline(self):
        self.assertEqual(self.f.readline(),
                         'A. HISTORY OF THE SOFTWARE')

        self.assertEqual(self.f.readline(),
                         '==========================')
        self.f.seek(0)

    def testIterateFile(self):
        self.assertEqual(self.f.bof(), True)
        self.assertEqual(self.f.eof(), False)
        for line in self.f:
            pass
        self.assertEqual(self.f.bof(), False)
        self.assertEqual(self.f.eof(), True)
        self.f.seek(0)
        
    def testReadparagraph(self):
        l = self.f.readparagraph()
        self.assertEqual(l, 'A. HISTORY OF THE SOFTWARE'+self.f.linesep+'==========================')
        l = self.f.readparagraph()
        self.assertEqual(l, 'Python was created in the early 1990s by Guido van Rossum at Stichting'+self.f.linesep+
                         'Mathematisch Centrum (CWI, see http://www.cwi.nl) in the Netherlands'+self.f.linesep+
                         'as a successor of a language called ABC.  Guido remains Python\'s'+self.f.linesep+
                         'principal author, although it includes many contributions from others.')
        self.f.cuepast("to make these releases possible.") # next paragraph is separated by three newlines
        t = self.f.readparagraph()[:23]
        self.assertEqual(t,"B. TERMS AND CONDITIONS")
        self.f.seek(0)

    def testReadChunk(self):
        l = self.f.readchunk('(')
        l = self.f.readchunk(')')
        self.assertEqual(l,'CWI, see http://www.cwi.nl')
        self.f.seek(0)

    def testPeekLine(self):
        l = self.f.peekline()
        self.assertEqual(l, 'A. HISTORY OF THE SOFTWARE')
        l = self.f.peekline(4)
        self.assertEqual(l, 'Python was created in the early 1990s by Guido van Rossum at Stichting')
        self.f.seek(0)
        
    def testPeekParagraph(self):
        l = self.f.peekparagraph()
        self.assertEqual(l, 'A. HISTORY OF THE SOFTWARE'+self.f.linesep+'==========================')
        l = self.f.peekparagraph(2)
        self.assertEqual(l, 'Python was created in the early 1990s by Guido van Rossum at Stichting'+self.f.linesep+
                         'Mathematisch Centrum (CWI, see http://www.cwi.nl) in the Netherlands'+self.f.linesep+
                         'as a successor of a language called ABC.  Guido remains Python\'s'+self.f.linesep+
                         'principal author, although it includes many contributions from others.')
        self.f.seek(0)

    

    def testPrevLine(self):
        self.f.readparagraph()
        self.f.readparagraph()
        self.assertEqual(self.f.prevline(3), # first two newlines, then the actual previous line (does this make sense?)
                         'principal author, although it includes many contributions from others.')
        self.assertEqual(self.f.prevline(6),
                         'Python was created in the early 1990s by Guido van Rossum at Stichting')
        self.f.seek(0)

    def testCue(self):
        self.f.cue("Guido")
        self.assertEqual(self.f.readline(),
                          'Guido van Rossum at Stichting')
        self.f.seek(0)

    def testCuePast(self):
        self.f.cuepast("Guido")
        self.assertEqual(self.f.readline(),
                          ' van Rossum at Stichting')
        self.f.seek(0)

    def testReadTo(self):
        self.assertEqual(self.f.readto("SOFTWARE"),
                          'A. HISTORY OF THE ')


# run all basic tests again, but this time initialised from a unicode buffer
class Ustring(Basic):
    def setUp(self):
        with codecs.open(PREFIX + "/LICENSE.txt",encoding='ascii') as fp:
            data = fp.read()
        self.f = TextReader(string=data,linesep=TextReader.UNIX)


class Codecs:
    def testUTF(self):
        f = TextReader(PREFIX + "/test/test_doctest4.txt", "utf-8")
        f.cue("u'f")
        self.assertEqual(f.read(5),
                          "u'f\u00f6\u00f6") 
        f.cue("u'b")
        self.assertEqual(f.read(5),
                          "u'b\u0105r")

    def testISO(self):
        f = TextReader(PREFIX + "/test/test_shlex.py", "iso-8859-1")
        f.cue(';|-|)|')
        f.readline()
        self.assertEqual(f.read(5),
                          "\u00e1\u00e9\u00ed\u00f3\u00fa")

    def testKOI8(self):
        f = TextReader(PREFIX + "/test/test_pep263.py", "koi8-r")
        f.cue('u"')
        self.assertEqual(f.read(7),
                          'u"\u041f\u0438\u0442\u043e\u043d')

class Processing(unittest.TestCase):
    def setUp(self):
        self.f = TextReader(PREFIX + "/LICENSE.txt",linesep=TextReader.UNIX)

    def testStrip(self):
        self.f.autostrip = True
        self.assertEqual(self.f.peekline(28),
                          'Release         Derived     Year        Owner       GPL-')
        self.f.autostrip = False
        self.assertEqual(self.f.peekline(28),
                          '    Release         Derived     Year        Owner       GPL-')
        self.f.seek(0)

    def testDewrap(self):
        self.f.autodewrap = True
        self.assertEqual(self.f.readparagraph(),
                          'A. HISTORY OF THE SOFTWARE ==========================')
        self.f.seek(0)
        self.f.autodewrap = False
        self.assertEqual(self.f.readparagraph(),
                          'A. HISTORY OF THE SOFTWARE'+self.f.linesep+'==========================')
        self.f.seek(0)
        

    def testDehyphenate(self):
        pass

    def testExpandtabs(self):
        pass

    def testReadTable(self):
        # Should this even be in the Processing test suite?
        pass
    

class Customiterator(unittest.TestCase):
    def setUp(self):
        self.f = TextReader(PREFIX + "/LICENSE.txt",linesep=TextReader.UNIX)

    def testIterateParagraph(self):
        cnt = 0
        for p in self.f.getiterator(self.f.readchunk,self.f.linesep*2):
            cnt += 1

        self.assertEqual(cnt, 44) 
        

class Subreaders(unittest.TestCase):
    def setUp(self):
        self.f = TextReader(PREFIX + "/test_base64.py",linesep=TextReader.UNIX)

    def testPage1(self):
        p = self.f.getreader(self.f.readpage)
        # print "p.maxpos: %s" % p.maxpos
        self.assertEqual(p.readline(),
                         'import unittest')
        self.assertRaises(IOError, p.peekline, 32) # we shouldn't be able to read ahead to page 2
        self.assertRaises(IOError, p.cue, 'LegacyBase64TestCase') # not by this method either
        self.f.seek(0)


    def testPage2(self):
        self.f.readpage() 
        p = self.f.getreader(self.f.readpage)
        p.readline()
        self.assertEqual(p.readline(),
                         'class LegacyBase64TestCase(unittest.TestCase):')

        self.assertRaises(IOError,p.prevline, 4) # we shouldn't be able to read backwards to page 1

        self.f.seek(0)
    


class Edgecases(unittest.TestCase):
    def setUp(self):
        self.f = TextReader(PREFIX + "/LICENSE.txt",linesep=TextReader.UNIX)

    def testPeekPastEOF(self):
        self.assertRaises(IOError,
                          self.f.peekline, 4711)

    def testPrevPastBOF(self):
        self.assertRaises(IOError,
                          self.f.prevline, 4711)

    def testReadPastEOF(self):
        self.assertEqual(len(self.f.read(1)), 1)
        self.f.read(sys.maxsize) # read past end of file - no license text is THAT big
        self.assertNotEqual(self.f.currpos, sys.maxsize+1)
        self.assertEqual(len(self.f.read(1)), 0) # no more to read
        self.assertEqual(len(self.f.readline()), 0)
        self.f.seek(0)
        
    def testReadlineUntilEOF(self):
        for line in self.f:
            prev = line
            pass
        self.assertEqual(prev,
                         'OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.')
        self.assertEqual(self.f.readline(), '')

    def testSearchInVain(self):
        self.assertRaises(IOError,
                          self.f.cue, 'I am a little teapot')
        self.f.seek(0)

class Fileops(unittest.TestCase):

    def testClose(self):
        f = open("textreader.tmp","w")
        f.write("foo")
        f.close()
        r = TextReader("textreader.tmp")
        # make sure TextReader isn't keeping the file open
        os.rename("textreader.tmp", "textreader2.tmp")
        os.unlink("textreader2.tmp")
        self.assertEqual("foo",r.readline())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
import logging
import re
import itertools
import codecs
from glob import glob
from bz2 import BZ2File

from lxml import etree
from six import text_type as str
# from six import binary_type as bytes
import six

from ferenda import util, errors
from .elements import UnicodeElement
from .elements import CompoundElement
from .elements import OrdinalElement


class PDFReader(CompoundElement):

    """Parses PDF files and makes the content available as a object
    hierarchy. After calling :py:meth:`~ferenda.PDFReader.read`, the
    PDFReader itself is a list of :py:class:`ferenda.pdfreader.Page`
    objects, which each is a list of
    :py:class:`ferenda.pdfreader.Textbox` objects, which each is a
    list of :py:class:`ferenda.pdfreader.Textelement` objects.

    .. note::

       This class depends on the command line tool pdftohtml from
       `poppler <http://poppler.freedesktop.org/>`_.

       The class can also handle any other type of document (such as
       Word/OOXML/WordPerfect/RTF) that OpenOffice or LibreOffice
       handles by first converting it to PDF using the ``soffice``
       command line tool (which then must be in your ``$PATH``).

       If the PDF contains only scanned pages (without any OCR
       information), the pages can be run through the ``tesseract``
       command line tool (which, again, needs to be in your
       ``$PATH``). You need to provide the main language of the
       document as the ``ocr_lang`` parameter, and you need to have
       installed the tesseract language files for that language.

    """

    tagname = "div"
    classname = "pdfreader"
    def __init__(self, *args, **kwargs):
        self.fontspec = {}
        self.log = logging.getLogger('pdfreader')
        if 'filename' in kwargs:
            self.filename = kwargs['filename']
        else:
            self.filename = None
        # super(PDFReader, self).__init__(*args, **kwargs)

    def read(self, pdffile, workdir, images=True, convert_to_pdf=False, keep_xml=True, ocr_lang=None):
        """Initializes a PDFReader object from an existing PDF file. After
        initialization, the PDFReader contains a list of
        :py:class:`~ferenda.pdfreader.Page` objects.

        :param pdffile: The full path to the PDF file (or, if
                        ``convert_to_pdf`` is set, any other document
                        file)
        :param workdir: A directory where intermediate files (particularly
                        background PNG files) are stored
        :param convert_to_pdf: If pdffile is any other type of
                               document other than PDF, attempt to
                               first convert it to PDF using the
                               ``soffice`` command line tool (from
                               OpenOffice/LibreOffice).
        :type  convert_to_pdf: bool
        :param keep_xml: If False, remove the intermediate XML
                         representation of the PDF that gets created
                         in ``workdir``. If true, keep it around to
                         speed up subsequent parsing operations. If
                         set to the special value ``"bz2"``, keep it
                         but compress it with :py:mod:`bz2`.
        :type  keep_xml: bool
        :param ocr_lang: If provided, PDFReader will extract scanned
                         images from the PDF file, and run an OCR
                         program on it, using the ``ocr_lang``
                         language heuristics. (Note that this is not
                         neccessarily an IETF language tag like "sv"
                         or "en-GB", but rather whatever the
                         underlying ``tesseract`` program uses).
        :param ocr_lang: str

        """
        # start by removing all pages left behind by a previous read
        self[:] = []
        
        if convert_to_pdf:
            newpdffile = workdir + os.sep + os.path.splitext(os.path.basename(pdffile))[0] + ".pdf"
            if not os.path.exists(newpdffile):
                util.ensure_dir(newpdffile)
                cmdline = "soffice --headless -convert-to pdf -outdir '%s' %s" % (workdir, pdffile)
                self.log.debug("%s: Converting to PDF: %s" % (pdffile, cmdline))
                (ret, stdout, stderr) = util.runcmd(
                    cmdline, require_success=True)
                pdffile = newpdffile
        
        self.filename = pdffile
        assert os.path.exists(pdffile), "PDF %s not found" % pdffile
        basename = os.path.basename(pdffile)
        stem = os.path.splitext(basename)[0]

        if ocr_lang:
            suffix = ".hocr.html"
            converter = self._tesseract
            converter_extra = {'lang': ocr_lang}
            parser = self._parse_hocr
        else:
            suffix = ".xml"
            converter = self._pdftohtml
            converter_extra = {'images': images}
            parser = self._parse_xml

        convertedfile = os.sep.join([workdir, stem + suffix])
        if keep_xml == "bz2":
            real_convertedfile = convertedfile + ".bz2"
        else:
            real_convertedfile = convertedfile

        
        tmppdffile = os.sep.join([workdir, basename])
        # copying the pdffile to the workdir is only needed if we use self._pdftohtml

        if not util.outfile_is_newer([pdffile], real_convertedfile):
            util.copy_if_different(pdffile, tmppdffile)
            # this is the expensive operation
            res = converter(tmppdffile, workdir, **converter_extra)
            if keep_xml == "bz2":
                with open(convertedfile, mode="rb") as rfp:
                    # BZ2File supports the with
                    # statement in py27+, but we
                    # support py2.6
                    wfp = BZ2File(real_convertedfile, "wb")
                    wfp.write(rfp.read())
                    wfp.close()
                os.unlink(convertedfile)
            else: # keep_xml = True
                pass

        if keep_xml == "bz2":
            # FIXME: explicitly state that encoding is utf-8 (in a py26 compatible manner
            fp = BZ2File(real_convertedfile)
        else:
            fp = codecs.open(real_convertedfile, encoding="utf-8")

        res = parser(fp, real_convertedfile)

        fp.close()
        if keep_xml == False:
            os.unlink(convertedfile)
        return res

    def is_empty(self):
        return 0 == sum([len(x) for x in self])

    def textboxes(self, gluefunc=None, pageobjects=False, keepempty=False):
        """Return an iterator of the textboxes available.

        ``gluefunc`` should be a callable that is called with
        (textbox, nextbox, prevbox), and returns True iff nextbox
        should be appended to textbox.

        If ``pageobjects``, the iterator can return Page objects to
        signal that pagebreak has ocurred (these Page objects may or
        may not have Textbox elements).

        If ``keepempty``, process and return textboxes that have no
        text content (these are filtered out by default)
        """
        textbox = None
        prevbox = None
        if gluefunc:
            glue = gluefunc
        else:
            glue = self._default_glue
        for page in self:
            if pageobjects:
                yield page
            for nextbox in page:
                if not (keepempty or str(nextbox).strip()):
                    continue
                if not textbox: # MUST glue
                    textbox = nextbox
                else:
                    if glue(textbox, nextbox, prevbox):
                        textbox += nextbox
                    else:
                        yield textbox
                        textbox = nextbox
                prevbox = nextbox
            if textbox:
                yield textbox
                textbox = None

    def drawboxes(self, outfile, gluefunc=None):
        """Create a copy of the parsed PDF file, but with the textboxes
        created by ``gluefunc`` clearly marked. Returns the name of
        the created pdf file.

        ..note::

          This requires PyPDF2 and reportlab, which aren't installed
          by default (and at least reportlab is not py3 compatible).

        """
        from PyPDF2 import PdfFileWriter, PdfFileReader
        from reportlab.pdfgen import canvas
        import StringIO

        packet = None
        output = PdfFileWriter()
        existing_pdf = PdfFileReader(open(self.filename, "rb"))
        pageidx = 0
        sf = 2/3.0 # scaling factor
        dirty = False
        for tb in self.textboxes(gluefunc, pageobjects=True):
            if isinstance(tb, Page):
                if dirty:
                    can.save()
                    packet.seek(0)
                    new_pdf = PdfFileReader(packet)
                    self.log.debug("Getting page %s from existing pdf" % pageidx)
                    page = existing_pdf.getPage(pageidx)
                    page.mergePage(new_pdf.getPage(0))
                    output.addPage(page)
                    pageidx += 1
                pagesize = (tb.width*sf, tb.height*sf)
                # print("pagesize %s x %s" % pagesize)
                packet = StringIO.StringIO()
                can = canvas.Canvas(packet, pagesize=pagesize,
                                    bottomup=False)
                can.setStrokeColorRGB(0.2,0.5,0.3)
                can.translate(0,0)
            else:
                dirty = True
                # x = repr(tb)
                # print(x)
                can.rect(tb.left*sf, tb.top*sf,
                         tb.width*sf, tb.height*sf)

        packet.seek(0)
        can.save()
        new_pdf = PdfFileReader(packet)
        self.log.debug("Getting last page %s from existing pdf" % pageidx)
        page = existing_pdf.getPage(pageidx)
        page.mergePage(new_pdf.getPage(0))
        output.addPage(page)
        outputStream = open(outfile, "wb")
        output.write(outputStream)
        outputStream.close()
        self.log.debug("wrote %s" % outfile)
        return outfile

    @staticmethod
    def _default_glue(textbox, nextbox, prevbox):
        # default logic: if lines are next to each other
        # horizontally, line up vertically, and have the same
        # font, then they should be glued
        linespacing = 1
        if (textbox.getfont() == nextbox.getfont() and
            textbox.left == nextbox.left and
            textbox.top + textbox.height + linespacing >= nextbox.top):
            return True


    def _tesseract(self, tmppdffile, workdir, lang):
        root = os.path.splitext(os.path.basename(tmppdffile))[0]

        # step 1: find the number of pages
        cmd = "pdfinfo %s" % tmppdffile
        (returncode, stdout, stderr) = util.runcmd(cmd, require_success=True)
        m = re.search("Pages:\s+(\d+)", stdout)
        number_of_pages = int(m.group(1))
        self.log.debug("%(root)s.pdf has %(number_of_pages)s pages" % locals())
        # step 2: extract the images (should be one per page), 10
        # pages at a time (pdfimages flakes out on larger loads)
        to_int = int
        for idx, i in enumerate(range(int(number_of_pages / 10) + 1)):
            frompage = (i * 10) + 1
            topage = min((i + 1) * 10, number_of_pages)
            if frompage > topage:
                continue
            cmd = "pdfimages -p -f %(frompage)s -l %(topage)s %(tmppdffile)s %(workdir)s/%(root)s" % locals()
            self.log.debug("- running "+cmd)
            (returncode, stdout, stderr) = util.runcmd(cmd, require_success=True)
            # step 2.1: Combine the recently extracted images and
            # into a new tif (so that we add 10
            # pages at a time to the tif, as imagemagick can
            # create a number of pretty large files for each page,
            # so converting 200 images will fill 10 G of your temp
            # space -- which we'd like to avoid)
            cmd = "convert %(workdir)s/%(root)s-*.pbm -compress Zip %(workdir)s/%(root)s-tmp%(idx)04d.tif" % locals()
            self.log.debug("- running " + cmd)
            (returncode, stdout, stderr) = util.runcmd(cmd, require_success=True)
            # step 2.2: Remove pbm files now that they're in the .tif
            for f in glob("%(workdir)s/%(root)s-*.pbm" % locals()):
                os.unlink(f)

        # Step 3: Combine all the 10-page tifs into a giant tif using tiffcp
        cmd = "tiffcp -c zip %(workdir)s/%(root)s-tmp*.tif %(workdir)s/%(root)s.tif" % locals()
        self.log.debug("- running " + cmd)
        (returncode, stdout, stderr) = util.runcmd(cmd, require_success=True)
        
                
        # Step 3: OCR the giant tif file to create a .hocr.html file
        # Note that -psm 1 (automatic page segmentation with
        # orientation and script detection) requires the installation
        # of tesseract-ocr-3.01.osd.tar.gz
        cmd = "tesseract %(workdir)s/%(root)s.tif %(workdir)s/%(root)s.hocr -l %(lang)s -psm 1 hocr" % locals()
        self.log.debug("running " + cmd)
        (returncode, stdout, stderr) = util.runcmd(cmd, require_success=True)

        # Step 5: Cleanup (the main .tif file can stay)
        os.unlink(tmppdffile)
        for f in glob("%(workdir)s/%(root)s-tmp*.tif" % locals()):
            os.unlink(f)

    def _pdftohtml(self, tmppdffile, workdir, images):
        root = os.path.splitext(os.path.basename(tmppdffile))[0]
        try:
            if images:
                # two pass coding: First use -c (complex) to extract
                # background pictures, then use -xml to get easy-to-parse
                # text with bounding boxes.
                cmd = "pdftohtml -nodrm -c %s" % tmppdffile
                self.log.debug("Converting: %s" % cmd)
                (returncode, stdout, stderr) = util.runcmd(cmd,
                                                          require_success=True)
                # we won't need the html files, or the blank PNG files
                for f in os.listdir(workdir):
                    if f.startswith(root) and f.endswith(".html"):
                        os.unlink(workdir + os.sep + f)
                    elif f.startswith(root) and f.endswith(".png"):
                        # this checks the number of unique colors in the
                        # bitmap. If there's only one color, we don't need
                        # the file
                        (returncode, stdout, stderr) = util.runcmd('convert %s -format "%%k" info:' % (workdir + os.sep + f))
                        if stdout.strip() == "1":
                            os.unlink(workdir + os.sep + f)
                        else:
                            self.log.debug("Keeping non-blank image %s" % f)

            # Without -fontfullname, all fonts are just reported as
            # having family="Times"...
            imgflag = "-i" if not images else ""

            cmd = "pdftohtml -nodrm -xml -fontfullname %s %s" % (imgflag, tmppdffile)
            self.log.debug("Converting: %s" % cmd)
            (returncode, stdout, stderr) = util.runcmd(cmd,
                                                       require_success=True)
            # if pdftohtml fails (if it's an old version that doesn't
            # support the fullfontname flag) it still uses returncode
            # 0! Only way to know if it failed is to inspect stderr
            # and look for if the xml file wasn't created.
            xmlfile = os.path.splitext(tmppdffile)[0] + ".xml"
            if stderr and not os.path.exists(xmlfile):
                raise errors.ExternalCommandError(stderr)


        finally:
            os.unlink(tmppdffile)
            assert not os.path.exists(tmppdffile)

    re_dimensions = re.compile("bbox (?P<left>\d+) (?P<top>\d+) (?P<right>\d+) (?P<bottom>\d+)").search
    def _parse_hocr(self, fp, filename):
        def dimensions(s):
            m = self.re_dimensions(s)
            return m.groupdict()
        tree = etree.parse(fp)
        for pageelement in tree.findall("//{http://www.w3.org/1999/xhtml}div[@class='ocr_page']"):
            dim = dimensions(pageelement.get('title'))
            page = Page(number=int(pageelement.get('id')[5:]),
                        width=int(dim['right']) - int(dim['left']),
                        height=int(dim['bottom']) - int(dim['top']),
                        background=None)
            # we discard elements at the ocr_carea (content area?)
            # level, we're only concerned with paragraph-level
            # elements
            for boxelement in pageelement.findall(".//{http://www.w3.org/1999/xhtml}p[@class='ocr_par']"):
                dim = dimensions(boxelement.get('title'))
                box = Textbox(top=int(dim['top']),
                              left=int(dim['left']),
                              width=int(dim['right']) - int(dim['left']),
                              height=int(dim['bottom']) - int(dim['top']),
                              font=None)
                for element in boxelement.findall(".//{http://www.w3.org/1999/xhtml}span[@class='ocrx_word']"):
                    dim = dimensions(element.get("title"))
                    t = "".join(element.itertext()) + element.tail
                    if element.getchildren():  # probably a <em> or <strong> element
                        tag = {'{http://www.w3.org/1999/xhtml}em': 'i',
                               '{http://www.w3.org/1999/xhtml}strong': 'b'}[element.getchildren()[0].tag]
                    else:
                        tag = None
                    text = Textelement(t,
                                       tag=tag,
                                       top=int(dim['top']),
                                       left=int(dim['left']),
                                       width=int(dim['right']) - int(dim['left']),
                                       height=int(dim['bottom']) - int(dim['top']),                    )
                    box.append(text)
                page.append(box)
            self.append(page)
        self.log.debug("PDFReader initialized: %d pages" %
                       (len(self)))
            
            
        

    def _parse_xml(self, xmlfp, xmlfilename):
        def txt(element_text):
            return re.sub(r"[\s\xa0\xc2]+", " ", str(element_text))
            
        self.log.debug("Loading %s" % xmlfilename)

        try:
            tree = etree.parse(xmlfp)
        except etree.XMLSyntaxError as e:
            self.log.warning("pdftohtml created incorrect markup, trying to fix using BeautifulSoup: %s" % e)
            xmlfp.seek(0)
            from bs4 import BeautifulSoup
            from io import BytesIO
            soup = BeautifulSoup(xmlfp, "xml")
            xmlfp = BytesIO(str(soup).encode("utf-8"))
            tree = etree.parse(xmlfp)
            self.log.debug("BeautifulSoup workaround successful")

        # for each page element
        for pageelement in tree.getroot():
            if pageelement.tag == "outline":
                # FIXME: we want to do something with this information
                continue
            page = Page(number=int(pageelement.attrib['number']),  # always int?
                        width=int(pageelement.attrib['width']),
                        height=int(pageelement.attrib['height']),
                        background=None)
            background = "%s%03d.png" % (
                os.path.splitext(xmlfilename)[0], page.number)

            # Reasons this file might not exist: it was blank and therefore removed, or We're running under RepoTester
            if os.path.exists(background):
                page.background = background

            assert pageelement.tag == "page", "Got <%s>, expected <page>" % page.tag
            for element in pageelement:
                if element.tag == 'fontspec':
                    fontid =  element.attrib['id']
                    # make sure we always deal with a basic dict (not
                    # lxml.etree._Attrib) where all keys are str
                    # object (not bytes)
                    self.fontspec[fontid] = dict([(k,str(v)) for k,v in element.attrib.items()])
                    if "+" in element.attrib['family']:
                        self.fontspec[fontid]['family'] = element.attrib['family'].split("+",1)[1]
                    
                elif element.tag == 'text':
                    # eliminate "empty" textboxes
                    if element.text and txt(element.text).strip() == "" and not element.getchildren():
                        # print "Skipping empty box"
                        continue

                    attribs = dict(element.attrib)
                    attribs['fontspec'] = self.fontspec
                    b = Textbox(**attribs)

                    if element.text and element.text.strip():
                        b.append(Textelement(txt(element.text), tag=None))
                    # The below loop could be done recursively to
                    # support arbitrarily deep nesting (if we change
                    # Textelement to be a non-unicode derived type),
                    # but pdftohtml should not create such XML (there
                    # is no such data in the PDF file)
                    for child in element:
                        grandchildren = child.getchildren()
                        # special handling of the <i><b> construct
                        if grandchildren != []:
                            # print "Grandchildren handling: %s '%s' '%s'" % (len(grandchildren),
                            #                                                child.text,
                            #                                                child.tail)
                            # Handle '<text><i><b>FordonsÃ¥r</b>            <b>Faktor</b> </i></text>'
                            # assert (len(grandchildren) == 1), "General grandchildren not supported"
                            
                            if child.text:
                                b.append(Textelement(txt(child.text), tag=child.tag))
                            b.append(Textelement(
                                txt(" ".join([x.text for x in grandchildren])), tag="ib"))
                            if child.tail:
                                b.append(Textelement(txt(child.tail), tag=None))
                        else:
                            b.append(
                                Textelement(txt(child.text), tag=child.tag))
                            if child.tail:
                                b.append(Textelement(txt(child.tail), tag=None))
                    if element.tail and element.tail.strip():  # can this happen?
                        b.append(Textelement(txt(element.tail), tag=None))
                    page.append(b)
            # done reading the page
            self.append(page)
        self.log.debug("PDFReader initialized: %d pages, %d fontspecs" %
                       (len(self), len(self.fontspec)))

    def median_box_width(self, threshold=0):
        """Returns the median box width of all pages."""
        boxwidths = []
        for page in self:
            for box in page:
                if box.right - box.left < threshold:
                    continue
                # print "Box width: %d" % (box.right-box.left)
                boxwidths.append(box.right - box.left)
        boxwidths.sort()
        return boxwidths[int(len(boxwidths) / 2)]

class Page(CompoundElement, OrdinalElement):

    """Represents a Page in a PDF file. Has *width* and *height* properties."""

    tagname = "div"
    classname = "pdfpage"

    @property
    def id(self):
        # FIXME: this will only work for documents consisting of a
        # single PDF file, not multiple (see
        # pdfdocumentrepository.create_external_resources to
        # understand why)
        return "page%03d" % self.number

    # text: can be string, re obj or callable (gets called with the box obj)
    # fontsize: can be int or callable
    # fontname: can be string or callable
    # top,left,bottom,right
    def boundingbox(self, top=0, left=0, bottom=None, right=None):
        """A generator of :py:class:`ferenda.pdfreader.Textbox` objects that
           fit into the bounding box specified by the parameters.

        """
        if not bottom:
            bottom = self.height
        if not right:
            right = self.width
        for box in self:
            if (box.top >= top and
                box.left >= left and
                box.bottom <= bottom and
                    box.right <= right):
                # print "    SUCCESS"
                yield box
            # else:
            #    print "    FAIL"

    def crop(self, top=0, left=0, bottom=None, right=None):
        """Removes any :py:class:`ferenda.pdfreader.Textbox` objects that does not fit within the bounding box specified by the parameters."""
        # Crop any text box that sticks out
        # Actually if top and left != 0, we need to adjust them
        newboxes = []
        for box in self.boundingbox(top, left, bottom, right):
            box.top = box.top - top
            box.left = box.left - left
            box.right = box.right - right
            box.bottom = box.bottom - bottom
            newboxes.append(box)
        self[:] = []
        self.extend(newboxes)
        self.width = right - left
        self.height = bottom - top
        # Then crop the background images... somehow
        if os.path.exists(self.background):
            cmdline = "convert %s -crop %dx%d+%d+%d +repage %s" % (self.background,
                                                                   self.width, self.height, left, top,
                                                                   self.background + ".new")
            # print "Running %s" % cmdline
            (returncode, stdout, stderr) = util.runcmd(cmdline,
                                                       require_success=True)
            util.replace_if_different(
                "%s.new" % self.background, self.background)

    def __str__(self):
        textexcerpt = " ".join([str(x) for x in self])
        return "Page %d (%d x %d): '%s...'" % (self.number, self.width, self.height, str(textexcerpt[:40]))

    def __repr__(self):
        return '<%s %d (%dx%d): %d textboxes>' % (self.__class__.__name__,
                                                   self.number, self.width, self.height,
                                                   len(self))

class Textbox(CompoundElement):

    """A textbox is a amount of text on a PDF page, with *top*, *left*,
*width* and *height* properties that specifies the bounding box of the
text. The *font* property specifies the id of font used (use
:py:meth:`~ferenda.pdfreader.Textbox.getfont` to get a dict of all
font properties). A textbox consists of a list of Textelements which
may differ in basic formatting (bold and or italics), but otherwise
all text in a Textbox has the same font and size.

    """
    tagname = "p"
    classname = "textbox"
    
    def __init__(self, *args, **kwargs):
        assert 'top' in kwargs, "top attribute missing"
        assert 'left' in kwargs, "left attribute missing"
        assert 'width' in kwargs, "width attribute missing"
        assert 'height' in kwargs, "height attribute missing"
        assert 'font' in kwargs, "font id attribute missing"

        self.top = int(kwargs['top'])
        self.left = int(kwargs['left'])
        self.width = int(kwargs['width'])
        self.height = int(kwargs['height'])
        self.right = self.left + self.width
        self.bottom = self.top + self.height

        # self.__fontspecid = kwargs['font']
        self.font = kwargs['font']
        if 'fontspec' in kwargs:
            self.__fontspec = kwargs['fontspec'] 
            del kwargs['fontspec']

        del kwargs['top']
        del kwargs['left']
        del kwargs['width']
        del kwargs['height']
        del kwargs['font']

        super(Textbox, self).__init__(*args, **kwargs)

    def __str__(self):
        return "".join(self)

    def __repr__(self):
        # <Textbox 30x18+278+257 "5.1">
        # <Textbox 430x14+287+315 "Regeringens fÃ¶rslag: NÃ¤[...]g ska ">
        s = str(self)
        if len(s) > 40:
            s = s[:25] + "[...]" + s[-10:]

        if six.PY2:
            # s = repr(s)
            s = s.encode('ascii', 'replace')
        if self.getfont():
            fontinfo = "%s@%s " % (self.getfont()['family'],
                                  self.getfont()['size'])
        else:
            fontinfo = ""
        return '<%s %sx%s+%s+%s %s"%s">' % (self.__class__.__name__,
                                            self.width, self.height,
                                            self.left, self.top,
                                            fontinfo,
                                            s)
    def __add__(self, other):
        # expand dimensions
        top = min(self.top, other.top)
        left = min(self.left, other.left)
        width = max(self.left + self.width,
                        other.left + other.width) - left
        height = max(self.top + self.height,
                          other.top + other.height) - top

        res = Textbox(top=top, left=left, width=width, height=height,
                      font=self.font,
                      fontspec=self.__fontspec)
        
        # add all text elements
        c = Textelement(tag=None)
        for e in itertools.chain(self, other):
            if e.tag != c.tag:
                res.append(c)
                c = Textelement(tag=e.tag)
            else:
                c = c + e
        res.append(c)
        return res


    def __iadd__(self, other):
        if len(self):
            c = self.pop()
        else:
            c = Textelement(tag=None)
        for e in other:
            if e.tag != c.tag:
                self.append(c)
                c = Textelement(tag=e.tag)
            else:
                c = c + e
        self.append(c)
        self.top = min(self.top, other.top)
        self.left = min(self.left, other.left)
        self.width = max(self.left + self.width,
                         other.left + other.width) - self.left
        self.height = max(self.top + self.height,
                          other.top + other.height) - self.top
        return self
        
    def as_xhtml(self, uri):
        element = super(Textbox, self).as_xhtml(uri)
        # FIXME: we should output these positioned style attributes
        # only when the resulting document is being serialized in a
        # positioned fashion. Possibly do some translation from PDF
        # points (which is what self.top, .left etc is using) and
        # pixels (which is what the CSS uses)
        element.set('style', 'top: %spx, left: %spx, height: %spx, width: %spx' % (self.top, self.left, self.height, self.width))
        return element


    def getfont(self):
        """Returns a fontspec dict of all properties of the font used."""
        if self.font:
            return self.__fontspec[self.font]
        else:
            return {}


class Textelement(UnicodeElement):
    """Represent a single part of text where each letter has the exact
    same formatting. The ``tag`` property specifies whether the text
    as a whole is bold (``'b'``) , italic(``'i'`` bold + italic
    (``'bi'``) or regular (``None``).
    """
    
    def _get_tagname(self):
        if self.tag:
            return self.tag
        else:
            return "span"

    tagname = property(_get_tagname)

    def __add__(self, other):
        return Textelement(str(self) + str(other), tag = self.tag)
    

# The below code fixes a error with incorrectly nested tags often
# found in pdftohtml generated xml. Main problem is that this relies
# on sgmllib which is not included in python3. This is commented out
# in the hope that more recent pdftohtml versions fix this problem in
# the right place

# import sgmllib
# from xml.sax.saxutils import escape as xml_escape
# import unicodedata
#
# class PDFXMLFix(sgmllib.SGMLParser):
#     selfclosing = ["fontspec"]
#
# preparations to remove invalid chars in handle_data
#     all_chars = (unichr(i) for i in range(0x10000))
#     control_chars = ''.join(
#         c for c in all_chars if unicodedata.category(c) == 'Cc')
# tab and newline are technically Control characters in
# unicode, but we want to keep them.
#     control_chars = control_chars.replace("\t", "").replace("\n", "")
#     control_char_re = re.compile('[%s]' % re.escape(control_chars))
#
#     def __init__(self):
#         sgmllib.SGMLParser.__init__(self)
#         self.tags = []
#         self.fp = None
#
#     def fix(self, filename):
#         usetempfile = not self.fp
#
#         if usetempfile:
#             tmpfile = mktemp()
#             self.fp = open(tmpfile, "w")
#
#         self.fp.write('<?xml version="1.0" encoding="UTF-8"?>')
#
#         f = open(filename)
#         while True:
#             s = f.read(8192)
#             if not s:
#                 break
#             self.feed(s)
#         self.close()
#
#         if usetempfile:
#             self.fp.close()
#             if util.replace_if_different(tmpfile, filename):
#                 print(("replaced %s with %s" % (filename, tmpfile)))
#             else:
#                 print(("%s was identical to %s" % (filename, tmpfile)))
#
#     def close(self):
#         sgmllib.SGMLParser.close(self)
#         if self.tags:
#             sys.stderr.write(
#                 "WARNING: opened tag(s) %s not closed" % self.tags)
#             self.fp.write(
#                 "".join(["</%s>" % x for x in reversed(self.tags)]))
#
#     def handle_decl(self, decl):
# self.fp.write "Decl: ", decl
#         self.fp.write("<!%s>" % decl)
#
#     def handle_data(self, data):
#         len_before = len(data)
#         data = xml_escape(self.control_char_re.sub('', data))
#         len_after = len(data)
# self.fp.write "Data: ", data.strip()
# if len_before != len_after:
# sys.stderr.write("WARNING: data changed from %s to %s chars: %r\n" % (len_before,len_after,data))
#         self.fp.write(data)
#
#     def unknown_starttag(self, start, attrs):
# self.fp.write "Start: ", start, attrs
#         if start in self.selfclosing:
#             close = "/"
#         else:
#             close = ""
#             self.tags.append(start)
# sys.stderr.write(repr(self.tags)+"\n")
#         if attrs:
#             fmt = ['%s="%s"' % (x[0], x[1]) for x in attrs]
#             self.fp.write("<%s %s%s>" % (start, " ".join(fmt), close))
#         else:
#             self.fp.write("<%s>" % start)
#
#     def unknown_endtag(self, end):
# sys.stderr.write(repr(self.tags)+"\n")
#         start = self.tags.pop()
#         if end != start and end in self.tags:
# sys.stderr.write("%s is not %s, working around\n" % (end, start))
#             self.fp.write("</%s>" % start)
#             self.fp.write("</%s>" % end)
#             self.fp.write("<%s>" % start)
#         else:
#             self.fp.write("</%s>" % end)



# -*- coding: utf-8 -*-
"""General ready-made grammars for use with
:py:class:`~ferenda.CitationParser`. See :doc:`../citationparsing` for
examples.

"""
from __future__ import unicode_literals
from pyparsing import *

#
#
# ferenda.citationpatterns.url
#
# Adapted from http://pyparsing.wikispaces.com/file/view/urlparse_.py/31853197/urlparse_.py
url_scheme = oneOf("http https ftp")
url_netloc = delimitedList(
    Word(printables, excludeChars="/."), ".", combine=True)
# To avoid matching trailing punctuation, eg "(see http://foo.org/)":
url_tokens_not_at_end = Word(",).")
url_path_tokens = Word(printables, excludeChars="?#,).")
url_path = delimitedList(url_path_tokens, url_tokens_not_at_end, combine=True)
url_query_tokens = Word(printables, excludeChars="#,).")
url_query = delimitedList(
    url_query_tokens, url_tokens_not_at_end, combine=True)
url_fragment_tokens = Word(printables, excludeChars=",).")
url_fragment = delimitedList(
    url_fragment_tokens, url_tokens_not_at_end, combine=True)
url = (url_scheme.setResultsName("scheme") + Literal("://").suppress() +
       url_netloc.setResultsName("netloc") +
       Optional(url_path).setResultsName("path") +
       Optional(Literal("?").suppress() + url_query).setResultsName("query") +
       Optional(Literal("#").suppress() + url_fragment).setResultsName("fragment")).setResultsName("url")
"""Matchs any URL like 'http://example.com/ or
'https://example.org/?key=value#fragment' (note: only the
schemes/protocols 'http', 'https' and 'ftp' are supported)"""

#
#
# ferenda.citationpatterns.eulaw
#
LongYear = Word(nums, exact=4)
ShortYear = Word(nums, exact=2)
Month = oneOf(
    "januari februari mars april maj juni juli augusti september oktober november december")
DayInMonth = Word(nums, max=2)
Date = (DayInMonth + Month + LongYear)

Ordinal = Word(nums).setResultsName("Ordinal")
Year = (ShortYear | LongYear).setResultsName("Year")
Association = oneOf("EG EEG").setResultsName("Association")

Institution = Literal('rådets') | Literal(
    'Europaparlamentets och rådets') | Literal('kommissionens')

LegalactType = oneOf("direktiv förordning").setResultsName("LegalactType")
Directive = Group(
    Year + "/" + Ordinal + "/" + Association).setResultsName("Directive")
Regulation = Group("(" + Association + ")" + "nr" + Ordinal + "/" + Year)

# "artikel 42.1" => Article: 42, Subarticle: 1
Article = "artikel" + Word(nums).setResultsName(
    "ArticleID") + Optional("." + Word(nums).setResultsName("SubarticleID"))
Legalact = Institution + LegalactType + (Directive | Regulation) + \
    "av den" + Date
ArticleLegalact = Article + "i" + Legalact

eulaw = MatchFirst(
    [ArticleLegalact, Legalact, Article]).setResultsName("EULegislation")
"""Matches EU Legislation references like 'direktiv 2007/42/EU'."""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rdflib import Graph

from ferenda.compat import unittest, patch


# SUT
from ferenda import testutil


class Main(unittest.TestCase):
    def setUp(self):
        class Tester(testutil.FerendaTestCase):
            def assertLessEqual(self, x,y,z): pass
            def assertEqual(self, want, got): pass
            def assertTrue(self, stmt): pass
            def fail(self, msg): return msg
        self.tester = Tester()
            
    def test_equalgraphs(self):
        msg = self.tester.assertEqualGraphs(
            Graph(identifier="a").parse(data="""
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix a: <http://example.org/actors/> .
a:nm0000102 a foaf:Person .
            """, format="turtle"),
            Graph(identifier="b").parse(data="""
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix a: <http://example.org/actors/> .
a:nm0000134 a foaf:Person .
            """, format="turtle"),
            exact=True)
        self.assertEqual("""<Graph identifier=a (<class 'rdflib.graph.Graph'>)> != <Graph identifier=b (<class 'rdflib.graph.Graph'>)>
1 unexpected triples were found
1 expected triples were not found
- <http://example.org/actors/nm0000102> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person>
+ <http://example.org/actors/nm0000134> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person>
""", msg)


    def test_equalxml(self):
        msg = self.tester.assertEqualXML(
            """<foo arg1='x' arg2='y'/>""",
            """<foo arg2='y' arg1='x'/>"""
            )
        self.assertEqual(None, msg)
        msg = self.tester.assertEqualXML(
            """<foo arg1='x' arg2='y'/>""",
            """<bar arg2='y' arg1='x'/>"""
            )
        self.assertEqual("""--- want.xml
+++ got.xml
@@ -1 +1 @@
-<foo arg1="x" arg2="y"></foo>
+<bar arg1="x" arg2="y"></bar>


ERRORS:Tags do not match: 'want': foo, 'got': bar""", msg)

        msg = self.tester.assertEqualXML(
            """<foo arg1='x' arg2='y'/>""",
            """<foo arg2='z' arg1='x'/>"""
            )
        self.assertEqual("""--- want.xml
+++ got.xml
@@ -1 +1 @@
-<foo arg1="x" arg2="y"></foo>
+<foo arg1="x" arg2="z"></foo>


ERRORS:Attributes do not match: 'want': arg2='y', 'got': arg2='z'""", msg)

        msg = self.tester.assertEqualXML(
            """<foo arg1='x'/>""",
            """<foo arg1='x' arg2='y'/>"""
            )
        self.assertEqual("""--- want.xml
+++ got.xml
@@ -1 +1 @@
-<foo arg1="x"></foo>
+<foo arg1="x" arg2="y"></foo>


ERRORS:'got' has an attribute 'want' is missing: arg2""", msg)
            



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
# This is a set of class methods, originally from
# ferenda.sources.legal.eu.EurlexTreaties, which perform graph
# analysis on documents from ferenda.sources.legal.eu.EurlexCaselaw.
# The main idea is to let them live in
# ferenda.sources.general.GraphAnalyze, generalize them, and make them
# available for any docrepo who wants their stuff analyzed (by
# inheriting from DocumentRepository and this class). For the time
# being though, the code should not be included in tests and coverage
# -- it's dead code, but with a hope of resuscitation.
import os
import subprocess
from datetime import datetime
from operator import itemgetter
from pprint import pprint

from lxml import etree as ET
from rdflib import Graph, Literal, BNode, URIRef, Collection, Namespace, RDF
from whoosh import analysis, fields, query, scoring
from whoosh.filedb.filestore import RamStorage, FileStorage

try:
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import numpy.array as na
    import networkx
    from networkx.algorithms.link_analysis.pagerank_alg import pagerank
    from networkx.algorithms.link_analysis.hits_alg import hits
except ImportError:
    plt = None
    cm = None
    na = None
    networkx = None
    pagerank = None
    hits = None

from ferenda import TripleStore
from ferenda import util

DCT = Namespace(util.ns['dct'])
RINFOEX = Namespace('http://lagen.nu/terms#')


class GraphAnalyze(object):

    def prep_annotation_file(self, basefile):
        goldstandard = self.eval_get_goldstandard(basefile)
        baseline_set = self.eval_get_ranked_set_baseline(basefile)
        baseline_map = self.eval_calc_map(
            self.eval_calc_aps(baseline_set, goldstandard))
        print("Baseline MAP %f" % baseline_map)
        self.log.info("Calculating ranked set (pagerank, unrestricted)")
        pagerank_set = self.eval_get_ranked_set(basefile, "pagerank",
                                                age_compensation=False,
                                                restrict_cited=False)
        pagerank_map = self.eval_calc_map(
            self.eval_calc_aps(pagerank_set, goldstandard))
        print("Pagerank MAP %f" % pagerank_map)
        sets = [{'label': 'Baseline',
                 'data': baseline_set},
                {'label': 'Gold standard',
                 'data': goldstandard},
                {'label': 'PageRank',
                 'data': pagerank_set}]

        g = Graph()
        g.bind('dct', self.ns['dct'])
        g.bind('rinfoex', self.ns['rinfoex'])

        XHT_NS = "{http://www.w3.org/1999/xhtml}"
        tree = ET.parse(self.parsed_path(basefile))
        els = tree.findall("//" + XHT_NS + "div")
        articles = []
        for el in els:
            if 'typeof' in el.attrib and el.attrib['typeof'] == "eurlex:Article":
                article = str(el.attrib['id'][1:])
                articles.append(article)
        for article in articles:
            self.log.info("Results for article %s" % article)
            articlenode = URIRef(
                "http://lagen.nu/ext/celex/12008E%03d" % int(article))
            resultsetcollectionnode = BNode()
            g.add((resultsetcollectionnode, RDF.type, RDF.List))
            rc = Collection(g, resultsetcollectionnode)
            g.add((articlenode, DCT["relation"], resultsetcollectionnode))
            for s in sets:
                resultsetnode = BNode()
                listnode = BNode()
                rc.append(resultsetnode)
                g.add((resultsetnode, RDF.type, RINFOEX[
                      "RelatedContentCollection"]))
                g.add((resultsetnode, DCT["title"], Literal(s["label"])))
                g.add((resultsetnode, DCT["hasPart"], listnode))
                c = Collection(g, listnode)
                g.add((listnode, RDF.type, RDF.List))
                if article in s['data']:
                    print(("    Set %s" % s['label']))
                    for result in s['data'][article]:
                        resnode = BNode()
                        g.add((resnode, DCT["references"], Literal(result[0])))
                        g.add((resnode, DCT["title"], Literal(result[1])))
                        c.append(resnode)
                        print(("        %s" % result[1]))

        return self.graph_to_annotation_file(g, basefile)

    def graph_to_image(self, graph, imageformat, filename):
        import pydot
        import rdflib
        dot = pydot.Dot()
        # dot.progs = {"dot": "c:/Program Files/Graphviz2.26.3/bin/dot.exe"}

        # code from rdflib.util.graph_to_dot, but adjusted to handle unicode
        nodes = {}
        for s, o in graph.subject_objects():
            for i in s, o:
                if i not in list(nodes.keys()):
                    if isinstance(i, rdflib.BNode):
                        nodes[i] = repr(i)[7:]
                    elif isinstance(i, rdflib.Literal):
                        nodes[i] = repr(i)[16:-1]
                    elif isinstance(i, rdflib.URIRef):
                        nodes[i] = repr(i)[22:-2]

        for s, p, o in graph.triples((None, None, None)):
            dot.add_edge(pydot.Edge(nodes[s], nodes[o], label=repr(p)[22:-2]))

        self.log.debug("Writing %s format to %s" % (imageformat, filename))
        util.ensure_dir(filename)
        dot.write(path=filename, prog="dot", format=imageformat)
        self.log.debug("Wrote %s" % filename)

    top_articles = []
    graph_filetype = "png"
    # yields an iterator of Article URIs

    def _articles(self, basefile):
        # Those articles we have gold standard sets for now
        self.top_articles = ['http://lagen.nu/ext/celex/12008E263',
                             'http://lagen.nu/ext/celex/12008E101',
                             'http://lagen.nu/ext/celex/12008E267',
                             'http://lagen.nu/ext/celex/12008E107',
                             'http://lagen.nu/ext/celex/12008E108',
                             'http://lagen.nu/ext/celex/12008E296',
                             'http://lagen.nu/ext/celex/12008E258',
                             'http://lagen.nu/ext/celex/12008E045',
                             'http://lagen.nu/ext/celex/12008E288',
                             'http://lagen.nu/ext/celex/12008E034',
                             ]

        # For evaluation, only return the 20 top cited articles (which
        # analyze_article_citations incidentally compute for us). For
        # full-scale generation, use commented-out code below.
        if not self.top_articles:
            self.top_articles = self.analyze_article_citations(quiet=True)
        return self.top_articles

        # For full-scale processing, return all articles present in e.g. TFEU:
        # XHT_NS = "{http://www.w3.org/1999/xhtml}"
        #tree = ET.parse(self.parsed_path(basefile))
        #els = tree.findall("//"+XHT_NS+"div")
        # for el in els:
        #    if 'typeof' in el.attrib and el.attrib['typeof'] == "eurlex:Article":
        #        yield el.attrib['about']

    # returns a RDFLib.Graph
    def _sameas(self):
        sameas = Graph()
        sameas_rdf = util.relpath(
            os.path.dirname(__file__) + "/../res/eut/sameas.n3")
        sameas.load(sameas_rdf, format="n3")
        return sameas

    def _query_cases(self, article, sameas):
        pred = util.ns['owl'] + "sameAs"
        q = ""
        if article:
            q += "{ ?subj eurlex:cites <%s> }\n" % article
            for equiv in sameas.objects(URIRef(article), URIRef(pred)):
                q += "    UNION { ?subj eurlex:cites <%s> }\n" % equiv

        return """
PREFIX eurlex:<http://lagen.nu/eurlex#>
PREFIX dct:<http://purl.org/dc/terms/>
SELECT DISTINCT ?subj WHERE {
    ?subj ?pred ?obj .
    %s
    FILTER (regex(str(?subj), "^http://lagen.nu/ext/celex/6"))
}
""" % (q)

    # Returns a python list of dicts
    def _query_cites(self, article, sameas, restrict_citing, restrict_cited, year=None):
        if not year:
            year = datetime.datetime.today().year
        pred = util.ns['owl'] + "sameAs"
        q = ""
        if restrict_citing:
            q += "{ ?subj eurlex:cites <%s> }\n" % article
            for equiv in sameas.objects(URIRef(article), URIRef(pred)):
                q += "    UNION { ?subj eurlex:cites <%s> }\n" % equiv

        if restrict_cited:
            if q:
                q += ".\n"
            q = "{?obj eurlex:cites <%s>}\n" % article
            for equiv in sameas.objects(URIRef(article), URIRef(pred)):
                q += "    UNION { ?obj eurlex:cites <%s> }\n" % equiv

        return """
PREFIX eurlex:<http://lagen.nu/eurlex#>
PREFIX dct:<http://purl.org/dc/terms/>
SELECT DISTINCT ?subj ?pred ?obj ?celexnum WHERE {
    ?subj ?pred ?obj .
    ?subj eurlex:celexnum ?celexnum.
    %s
    FILTER (regex(str(?obj), "^http://lagen.nu/ext/celex/6") &&
            ?pred = eurlex:cites &&
            str(?celexnum) < str("6%s"@en))
}
""" % (q, year)

    def temp_analyze(self):
        store = TripleStore(self.config.storetype,
                            self.config.storelocation,
                            self.config.storerepository)
        # sq = self._query_cites('http://lagen.nu/ext/celex/12008E045',self._sameas(),False, True, 2012)
        sq = self._query_cites(None, self._sameas(), False, False, 2012)
        print(sq)
        cites = store.select(sq, format="python")
        self.log.debug(
            "    Citation graph contains %s citations" % (len(cites)))

        # remove duplicate citations, self-citations and pinpoints
        # in citations
        citedict = {}
        for cite in cites:
            # print repr(cite)
            if "-" in cite['obj']:
                cite['obj'] = cite['obj'].split("-")[0]

            if (cite['subj'] != cite['obj']):
                citedict[(cite['subj'], cite['obj'])] = True

        self.log.debug(
            "    Normalized graph contains %s citations" % len(citedict))

        degree = {}
        for citing, cited in list(citedict.keys()):
            if citing not in degree:
                degree[citing] = []
            if cited not in degree:
                degree[cited] = []
            degree[cited].append(citing)

        return

    def analyze(self):
        articles = self.analyze_article_citations(num_of_articles=10)
        # articles = self._articles('tfeu')
        self.analyze_baseline_queries(articles)
        self.analyze_citation_graphs(articles)

    def analyze_article_citations(self, num_of_articles=20, quiet=False):
        """Prints and returns a list of the top 20 most important articles in the
        TFEU treaty, as determined by the number of citing cases."""

        # Create a mapping of article equivalencies, eg Art 28 TEC == Art 34 TFEU
        sameas = self._sameas()
        equivs = {}
        pred = util.ns['owl'] + "sameAs"
        for (s, o) in sameas.subject_objects(URIRef(pred)):
            equivs[str(o)] = str(s)
        self.log.debug(
            "Defined %s equivalent article references" % len(equivs))

        # Select unique articles citings
        store = TripleStore(self.config.storetype,
                            self.config.storelocation,
                            self.config.storerepository)
        sq = """PREFIX eurlex:<http://lagen.nu/eurlex#>
                SELECT DISTINCT ?case ?article WHERE {
                    ?case eurlex:cites ?article .
                    FILTER (regex(str(?article), "^http://lagen.nu/ext/celex/1"))
             }"""
        cites = store.select(sq, format="python")

        citationcount = {}
        unmapped = {}
        self.log.debug("Going through %s unique citations" % len(cites))
        for cite in cites:
            article = cite['article'].split("-")[0]
            if "12008M" in article:
                pass
            elif article in equivs:
                article = equivs[article]
            else:
                if article in unmapped:
                    unmapped[article] += 1
                else:
                    unmapped[article] = 1
                article = None

            # Keep track of the number of citing cases
            if article:
                if article in citationcount:
                    citationcount[article] += 1
                else:
                    citationcount[article] = 1

        # Report the most common cites to older treaty articles that
        # we have no equivalents for in TFEU
        # sorted_unmapped = sorted(unmapped.iteritems(), key=itemgetter(1))[-num_of_articles:]
        # if not quiet:
        #    print "UNMAPPED:"
        #    pprint(sorted_unmapped)

        # Report and return the most cited articles
        sorted_citationcount = sorted(iter(list(
            citationcount.items())), key=itemgetter(1))[-num_of_articles:]
        if not quiet:
            print("CITATION COUNTS:")
            pprint(sorted_citationcount)
        return [x[0] for x in reversed(sorted_citationcount)]

    def analyze_baseline_queries(self, analyzed_articles, num_of_keyterms=5):
        basefile = "tfeu"
        # Helper from http://effbot.org/zone/element-lib.htm

        def flatten(elem, include_tail=0):
            text = elem.text or ""
            for e in elem:
                text += flatten(e, 1)
                if include_tail and elem.tail:
                    text += elem.tail
            return text
        # step 1: Create a temporary whoosh index in order to find out
        # the most significant words for each article

        #ana = analysis.StandardAnalyzer()
        ana = analysis.StemmingAnalyzer()
        # vectorformat = formats.Frequency(ana)
        schema = fields.Schema(article=fields.ID(unique=True),
                               content=fields.TEXT(analyzer=ana,
                                                   stored=True))

        st = RamStorage()
        tmpidx = st.create_index(schema)
        w = tmpidx.writer()

        XHT_NS = "{http://www.w3.org/1999/xhtml}"
        tree = ET.parse(self.parsed_path(basefile))
        els = tree.findall("//" + XHT_NS + "div")
        articles = []
        for el in els:
            if 'typeof' in el.attrib and el.attrib['typeof'] == "eurlex:Article":
                text = util.normalize_space(flatten(el))
                article = str(el.attrib['about'])
                articles.append(article)
                w.update_document(article=article, content=text)
        w.commit()
        self.log.info("Indexed %d articles" % len(articles))

        # Step 2: Open the large whoosh index containing the text of
        # all cases. Then, for each article, use the 5 most distinctive terms
        # (filtering away numbers) to create a query against that index
        tempsearch = tmpidx.searcher()
        g = Graph()
        g.bind('celex', 'http://lagen.nu/ext/celex/')
        g.bind('ir', 'http://lagen.nu/informationretrieval#')
        IR = Namespace('http://lagen.nu/informationretrieval#')
        # celex:12008E264 ir:keyterm "blahonga"@en.

        outfile = self.generic_path("keyterms", "analyzed", ".tex")
        util.ensure_dir(outfile)
        fp = open(outfile, "w")
        fp.write("""
\\begin{tabular}{r|%s}
  \\hline
  \\textbf{Art.} & \\multicolumn{%s}{l}{\\textbf{Terms}} \\\\
  \\hline
""" % ("l" * num_of_keyterms, num_of_keyterms))

        for article in analyzed_articles:
            fp.write(str(int(article.split("E")[1])))
            r = tempsearch.search(query.Term("article", article))
            terms = r.key_terms("content", numterms=num_of_keyterms + 1)
            terms = [t[0] for t in terms if not t[0].isdigit(
            )][:num_of_keyterms]
            for term in terms:
                fp.write(" & " + term)
                g.add((
                    URIRef(article), IR["keyterm"], Literal(term, lang="en")))
            self.log.debug("Article %s:%r" % (article, terms))
            fp.write("\\\\\n")
        fp.write("""
  \\hline
\\end{tabular}
""")
        fp.close()

        outfile = self.generic_path("keyterms", "analyzed", ".n3")
        util.ensure_dir(outfile)
        fp = open(outfile, "w")
        fp.write(g.serialize(format="n3"))
        fp.close()

    def analyze_citation_graphs(self, articles=None):
        # Basic setup
        # articles = self._articles('tfeu')[-1:]
        if not articles:
            articles = [None]
        if None not in articles:
            articles.append(None)
        this_year = datetime.datetime.today().year
        store = TripleStore(self.config.storetype,
                            self.config.storelocation,
                            self.config.storerepository)
        sameas = self._sameas()
        distributions = []

        # For each article (and also for no article = the entire citation graph)
        for article in articles:
            # Get a list of all eligble cases (needed for proper degree distribution)
            sq = self._query_cases(article, sameas)
            # print sq
            cases = {}
            caserows = store.select(sq, format="python")
            for r in caserows:
                cases[r['subj']] = 0

            self.log.info(
                "Creating graphs for %s (%s cases)" % (article, len(cases)))
            # Step 1. SPARQL the graph on the form ?citing ?cited
            # (optionally restricting on citing a particular article)
            if article:
                sq = self._query_cites(
                    article, sameas, True, False, this_year + 1)
            else:
                sq = self._query_cites(
                    None, sameas, False, False, this_year + 1)

            cites = store.select(sq, format="python")
            self.log.debug(
                "    Citation graph contains %s citations" % (len(cites)))

            # remove duplicate citations, self-citations and pinpoints
            # in citations
            citedict = {}
            missingcases = {}
            for cite in cites:
                # print repr(cite)
                if "-" in cite['obj']:
                    cite['obj'] = cite['obj'].split("-")[0]

                if not cite['obj'] in cases:
                    # print "Case %s (cited in %s) does not exist!\n" % (cite['obj'], cite['subj'])
                    missingcases[cite['obj']] = True
                    continue

                if (cite['subj'] != cite['obj']):
                    citedict[(cite['subj'], cite['obj'])] = True

            self.log.debug(
                "    Normalized graph contains %s citations (%s cited cases not found)" %
                (len(citedict), len(missingcases)))
            # pprint(missingcases.keys()[:10])

            # Step 2. Dotify the list (maybe the direction of arrows from
            # cited to citing can improve results?) to create a citation
            # graph
            self.analyse_citegraph_graphviz(list(citedict.keys()), article)

            # Step 3. Create a degree distribution plot
            degree, distribution = self.analyze_citegraph_degree_distribution(
                cases, list(citedict.keys()), article)
            if article:
                distributions.append([article, distribution])

            # Step 4. Create a citation/age scatterplot (or rather hexbin)
            self.analyze_citegraph_citation_age_plot(
                list(citedict.keys()), degree, distribution, article)

        # Step 5. Create a combined degree distribution graph of the
        # distinct citation networks. Also add the degree distribution
        # of gold standard cases

        self.analyze_citegraph_combined_degree_distribution(distributions)

    def analyse_citegraph_graphviz(self, cites, article, generate_graph=False):
        """Create a dot file (that can later be processed with dot or gephi)"""
        from time import time

        filetype = self.graph_filetype
        if article:
            filename = "citegraph_%s" % article.split("/")[-1]
        else:
            filename = "citegraph_all"

        dot_filename = self.generic_path(filename, "analyzed", ".dot")
        self.log.debug("    Writing graphwiz citation graph for %s" % article)
        fp = open(dot_filename, "w")
        fp.write("""digraph G {
                    graph [
                          ];
""")
        cnt = 0
        for citing, cited in cites:
            cnt += 1
            citing = citing.split("/")[-1]
            cited = cited.split("/")[-1]
            try:
                fp.write("  \"%s\" -> \"%s\" ;\n" % (citing, cited))
            except:
                pass
        fp.write("}")
        fp.close()

        if generate_graph:
            graph_filename = self.generic_path(
                dot_filename, "analyzed", "." + filetype)
            engine = "dot"
            start = time()
            cmdline = "%s -T%s -o%s tmp.dot" % (
                engine, filetype, graph_filename)
            self.log.debug("Running %s" % cmdline)
            p = subprocess.Popen(cmdline, shell=True)
            p.wait()
            self.log.info("Graph %s created in %.3f sec" % (
                graph_filename, time() - start))

    def analyze_citegraph_degree_distribution(self, cases, cites, article):
        self.log.debug("    Writing degree distribution graph")
        degree = cases
        # self.log.debug("    %s cases, first elements %r" % (len(cases),cases.values()[:5]))
        # this_year = datetime.datetime.today().year
        maxcites = 40
        # maxage = this_year - 1954

        for citing, cited in cites:
            if citing not in degree:
                degree[citing] = 0
            if cited not in degree:
                degree[cited] = 0
            degree[cited] += 1

        distribution = [0] * (max(degree.values()) + 1)

        for value in list(degree.values()):
            distribution[value] += 1

        fig = plt.figure()
        fig.set_size_inches(8, 4)
        ax = plt.subplot(111)
        ax.set_ylabel('Number of cases being cited <x> times')
        ax.set_xlabel('Number of citing cases (max %s)' % maxcites)
        ax.set_title('Degree distribution of case citations')

        filetype = self.graph_filetype
        if article:
            filename = "degree_distribution_%s" % (article.split("/")[-1])
        else:
            filename = "degree_distribution_all"
        filename = self.generic_path(filename, "analyzed", "." + filetype)

        plt.plot(distribution[:maxcites])
        plt.savefig(filename)
        plt.close()
        self.log.debug("    Created %s" % filename)
        return (degree, distribution)

    def analyze_citegraph_combined_degree_distribution(self, distributions):
        self.log.debug("    Writing combined degree distribution graph")
        # this_year = datetime.datetime.today().year
        maxcites = 40
        # maxnumber = 1000
        # maxage = this_year - 1954

        fig = plt.figure()
        fig.set_size_inches(8, 4)
        ax = plt.subplot(111)
        ax.set_ylabel('Number of cases being cited <x> times')
        ax.set_xlabel('Number of citing cases (max %s)' % maxcites)
        ax.set_title('Degree distribution of case citations concering specific articles')

        filetype = self.graph_filetype
        filename = "degree_distribution_combined"
        filename = self.generic_path(filename, "analyzed", "." + filetype)

        styles = []
        for i in range(1, 5):
            for j in (['-', '--', '-.', ':']):
            # for j in (['-','-','-','-','-']):
                styles.append((i, j))

        cnt = 0
        for (article, distribution) in distributions:
            label = article.split("/")[-1].split("E")[1]
            self.log.debug(
                "        Plotting %s %r" % (label, distribution[:4]))
            if label.isdigit():
                label = "Art. %s" % int(label)
            # label += " (%s uncited)" % distribution[0]
            lw, ls = styles[cnt]
            plt.plot(distribution[:maxcites], label=label,
                     linestyle=ls, linewidth=lw)

        # plt.axis([0,maxcites,0,maxnumber])
        plt.legend(loc='best',
                   markerscale=4,
                   prop={'size': 'x-small'},
                   ncol=int(len(distributions) / 6) + 1)

        plt.savefig(filename)
        plt.close()
        self.log.debug("    Created %s" % filename)

    def analyze_citegraph_citation_age_plot(self, cites, degree, distribution, article):
        self.log.debug("    Writing citation age plot")
        this_year = datetime.datetime.today().year
        maxcites = 40
        maxage = this_year - 1954

        cited_by_age = []
        citations = []
        for case in sorted(degree.keys()):
            try:
                year = int(case[27:31])
                caseage = this_year - year
                if year < 1954:
                    continue
            except ValueError:
                # some malformed URIs/Celexnos
                continue
            if degree[case] <= maxcites:
                cited_by_age.append(caseage)
                citations.append(degree[case])

        cases_by_age = [0] * (maxage + 1)
        for citing, cited in cites:
            year = int(citing[27:31])
            caseage = this_year - year
            if year < 1954:
                continue
            if caseage < 0:
                continue
            cases_by_age[caseage] += 1

        fig = plt.figure()
        fig.set_size_inches(8, 5)
        plt.axis([0, maxage, 0, maxcites])
        ax = plt.subplot(211)
        plt.hexbin(cited_by_age, citations, gridsize=maxcites,
                   bins='log', cmap=cm.hot_r)
        # plt.scatter(age,citations)
        ax.set_title("Distribution of citations by age")
        ax.set_ylabel("# of citations")
        #cb = plt.colorbar()
        # cb.set_label('log(# of cases with # of citations)')
        ax = plt.subplot(212)
        ax.set_title("Distribution of cases by age")
        plt.axis([0, maxage, 0, max(cases_by_age)])
        plt.bar(na.array(list(range(len(cases_by_age)))) + 0.5, cases_by_age)

        filetype = self.graph_filetype
        if article:
            filename = "citation_age_plot_%s" % (article.split("/")[-1])
        else:
            filename = "citation_age_plot_all"
        filename = self.generic_path(filename, "analyzed", "." + filetype)

        plt.savefig(filename)
        plt.close()
        self.log.debug("    Created %s" % filename)

#
# Evaluation

    def evaluate(self):
        result_cache = self.generic_path("result_cache", "eval", ".py")
        if os.path.exists(result_cache):
        # if False:
            self.log.info("Using result cache in %s" % result_cache)
            sets = eval(open(result_cache).read())
        else:
            sets = (
                ('baseline', self.eval_get_ranked_set_baseline('tfeu')),
                ('indegree_uncomp_unrestr', self.eval_get_ranked_set(
                    'tfeu', 'indegree', False, False)),
                ('indegree_uncomp_restr', self.eval_get_ranked_set(
                    'tfeu', 'indegree', False, True)),
                ('indegree_comp_unrestr', self.eval_get_ranked_set(
                    'tfeu', 'indegree', True, False)),
                ('indegree_comp_restr',
                 self.eval_get_ranked_set('tfeu', 'indegree', True, True)),
                ('hits_uncomp_unrestr',
                 self.eval_get_ranked_set('tfeu', 'hits', False, False)),
                ('hits_uncomp_restr',
                 self.eval_get_ranked_set('tfeu', 'hits', False, True)),
                ('hits_comp_unrestr',
                 self.eval_get_ranked_set('tfeu', 'hits', True, False)),
                ('hits_comp_restr',
                 self.eval_get_ranked_set('tfeu', 'hits', True, True)),
                ('pagerank_uncomp_unrestr', self.eval_get_ranked_set(
                    'tfeu', 'pagerank', False, False)),
                ('pagerank_uncomp_restr', self.eval_get_ranked_set(
                    'tfeu', 'pagerank', False, True)),
                ('pagerank_comp_unrestr', self.eval_get_ranked_set(
                    'tfeu', 'pagerank', True, False)),
                ('pagerank_comp_restr',
                 self.eval_get_ranked_set('tfeu', 'pagerank', True, True)),
            )
            util.ensure_dir(result_cache)
            fp = open(result_cache, "w")
            pprint(sets, fp)
            fp.close()

        aps_cache = self.generic_path("aps_cache", "eval", ".py")
        if os.path.exists(aps_cache):
        # if False:
            self.log.info("Using avg precision cache in %s" % aps_cache)
            avg_precisions = eval(open(aps_cache).read())
        else:
            goldstandard = self.eval_get_goldstandard('tfeu')
            avg_precisions = []
            for label, rankedset in sets:
                aps = self.eval_calc_aps(rankedset, goldstandard)
                avg_precisions.append((label, aps))

            fp = open(aps_cache, "w")
            pprint(avg_precisions, fp)
            fp.close()

        self.eval_aps_table(avg_precisions)

        if len(avg_precisions) > 5:
            maps = []
            for label, aps in avg_precisions:
                maps.append(self.eval_calc_map(aps))
            maps.sort(reverse=True)
            thresh = maps[5]
        else:
            thresh = 0.0

        top_avg_precisions = []
        for label, aps in avg_precisions:
            map_ = self.eval_calc_map(aps)
            self.log.info("%25s: MAP %s" % (label, map_))
            if (map_ > thresh) or (label == 'baseline'):
                top_avg_precisions.append(
                    ("%s: MAP %.3f" % (label, map_), aps))

        self.eval_aps_chart(top_avg_precisions)

    def eval_calc_aps(self, rankedset, goldstandard):
        """Calculate a set of average precisions for the given set of
        result sets for some information needs, compared to the gold
        standard for those information needs.

        Both rankedset and goldstandard are dicts with lists as values."""

        aps = []
        for infoneed in list(goldstandard.keys()):
            relevants = goldstandard[infoneed]
            if relevants:
                #self.log.debug("  Calculating AP for %s" % infoneed)
                pass
            else:
                self.log.debug(
                    "   No AP for %s: no gold standard" % (infoneed))
                continue

            ranking = rankedset[infoneed]
            precisions = []

            # for each relevant doc in the gold standard, check at
            # what position in the ranking the doc occurrs. Check the
            # precision of the ranking up to and including that position
            for relevant in relevants:
                try:
                    place = ranking.index(relevant)
                    relevant_cnt = 0
                    for r in ranking[:place + 1]:
                        if r in relevants:
                            relevant_cnt += 1
                    precision = float(relevant_cnt) / float(place + 1)
                    #self.log.debug("    Relevant result %s found at %s (relevant_cnt %s), precision %s" % (relevant.split("/")[-1], place+1,relevant_cnt,precision))
                except ValueError:
                    #self.log.debug("    Relevant result %s not found, precision: 0" % relevant.split("/")[-1])
                    precision = 0

                precisions.append(precision)

            ap = sum(precisions) / float(len(precisions))
            self.log.info("   AP for %s: %s" % (infoneed, ap))
            aps.append(ap)

        return aps

    def eval_calc_map(self, average_precision_set):
        return sum(average_precision_set) / float(len(average_precision_set))

    _graph_cache = {}

    def eval_get_ranked_set(self, basefile, algorithm="pagerank",
                            age_compensation=False, restrict_cited=True):
        # * algorithm: can be "indegree", "hits" or "pagerank".
        # * age_compensation: create one graph per year and average to
        #   compensate for newer cases (that have had less time to gain
        #   citations)
        # * restrict_cited: Use only such citations that exist between
        #   two cases that both cite the same TFEU article (othewise,
        #   use all citations from all cases that cite the TFEU
        #   article, regardless of whether the cited case also cites
        #   the same TFEU article)
        sameas = self._sameas()
        store = TripleStore(self.config.storetype,
                            self.config.storelocation,
                            self.config.storerepository)
        res = {}

        self.log.debug("Creating ranked set (%s,age_compensation=%s,restrict_cited=%s)" %
                       (algorithm, age_compensation, restrict_cited))

        for article in self._articles(basefile):
            article_celex = article.split("/")[-1]
            self.log.debug("    Creating ranking for %s" % (article_celex))
            this_year = datetime.datetime.today().year
            if age_compensation:
                years = list(range(1954, this_year + 1))
                # years = range(this_year-3,this_year) # testing
            else:
                years = list(range(this_year, this_year + 1))

            result_by_years = []
            for year in years:
                restrict_citing = True  # always performs better
                if (article, year, restrict_cited) in self._graph_cache:
                    # self.log.debug("Resuing cached graph (%s) for %s in %s" %
                    #               (restrict_cited, article_celex,year))
                    graph = self._graph_cache[(article, year, restrict_cited)]
                else:
                    # self.log.debug("Calculating graph for %s in %s" %
                    #               (article_celex,year))
                    sq = self._query_cites(article, sameas, restrict_citing,
                                           restrict_cited, year)
                    links = store.select(sq, format="python")
                    graph = self.eval_build_nx_graph(links)
                    self._graph_cache[(article, year, restrict_cited)] = graph
                    self.log.debug("      Citegraph for %s in %s has %s edges, %s nodes" %
                                   (article_celex, year, len(graph.edges()),
                                    len(graph.nodes())))

                if len(graph.nodes()) == 0:
                    continue

                ranked = self.eval_rank_graph(graph, algorithm)
                result_by_years.append({})
                for result, score in ranked:
                    result_by_years[-1][result] = score

            if age_compensation:
                compensated_ranking = {}
                for d, score in ranked:  # the last result set
                    # cut out the year part of the URI
                    celex = d.split("/")[-1]
                    try:
                        age = this_year + 1 - int(
                            celex[1:5])  # cases decided this year has age 1
                        # scores = [0,0,0 ... 3,4,8,22]
                        scores = [result_by_year[d]
                                  for result_by_year
                                  in result_by_years
                                  if d in result_by_year]
                        avg_score = sum(scores) / float(age)
                        # self.log.debug("Result %s (age %s, avg score %s) %r" %
                        #               (d,age,avg_score,scores))
                        compensated_ranking[d] = avg_score
                    except ValueError:
                        continue

            # return just a list of results, no scores
            if age_compensation:
                res[article] = [result for result in sorted(
                    compensated_ranking, key=compensated_ranking.__getitem__, reverse=True)]
            else:
                res[article] = [result[0] for result in ranked]

        return res

    def eval_build_nx_graph(self, links):
        #self.log.debug("Building graph with %s links" % len(links))
        nxgraph = networkx.DiGraph()
        for link in links:
            if "-" in link['obj']:
                nxgraph.add_edge(link['subj'], link['obj'].split("-")[0])
            else:
                nxgraph.add_edge(link['subj'], link['obj'])
        #self.log.debug("Graph has %s nodes" % len (nxgraph.nodes()))
        return nxgraph

    def eval_rank_graph(self, graph, algorithm="pagerank"):
        # should return a list of tuples (result,score) sorted in
        # reversed order (ie highest score first)
        if algorithm == "pagerank":
            ranked = pagerank(graph)
        elif algorithm == "hits":
            ranked = hits(graph, max_iter=10000)[1]  # 0: hubs, 1: authorities
        elif algorithm == "indegree":
            ranked = graph.in_degree()
        else:
            self.log.error(
                "Unknown ranking algorithm %s specified" % algorithm)
        sortedrank = sorted(
            iter(list(ranked.items())), key=itemgetter(1), reverse=True)
        return sortedrank

    # computes a ranked set for each baseline using a naive search
    # (using the most significant words of each article) and the
    # standard BM25F ranking function
    def eval_get_ranked_set_baseline(self, basefile):
        # Step 1: Read the saved keyterms for a subset of articles
        # (created by analyze_baseline_queries)
        g = Graph()
        g.parse(self.generic_path("keyterms", "analyzed", ".n3"), format="n3")

        articles = {}
        for (s, p, o) in g:
            if not str(s) in articles:
                articles[str(s)] = []
            articles[str(s)].append(str(o))

        # Step 2: Open the large whoosh index containing the text of
        # all cases. Then, create a query for each article based on
        # the keyterms.
        connector = query.Or
        indexdir = os.path.sep.join([self.config.datadir, 'ecj', 'index'])
        storage = FileStorage(indexdir)
        idx = storage.open_index()
        searcher = idx.searcher(weighting=scoring.BM25F())

        res = {}

        # for article in sorted(articles.keys()):
        for article in self._articles(basefile):
            terms = articles[article]
            rankedset = []
            #parser = qparser.QueryParser("content", idx.schema)
            #q = parser.parse(connector.join(terms))
            q = query.And([
                # query.Term("articles", article),
                connector([query.Term("content", x) for x in terms])
            ])
            # print q
            # self.log.debug("Article %s: %s", article, " or ".join(terms))
            results = searcher.search(q, limit=None)
            resultidx = 0
            # self.log.info("Keyterms for result: %r" % results.key_terms("content", docs=10, numterms=10))
            for result in results:
                reslbl = "%s (%s)" % (
                    result['basefile'], results.score(resultidx))
                rankedset.append([result['basefile'], reslbl])
                # self.log.debug(u"\t%s: %2.2d" % (result['title'], results.score(resultidx)))
                resultidx += 1
            self.log.info("Created baseline ranked set for %s: Top result %s (of %s)" %
                          (article.split("/")[-1], rankedset[0][0], len(rankedset)))

            # return just a list of URIs, no scoring information. But the
            # full URI isnt available in the whoosh db, so we recreate it.
            res[article] = ["http://lagen.nu/ext/celex/%s" % x[
                0] for x in rankedset]

        return res

    def eval_get_goldstandard(self, basefile):
        goldstandard = Graph()
        goldstandard_rdf = util.relpath(
            os.path.dirname(__file__) + "/../res/eut/goldstandard.n3")
        goldstandard.load(goldstandard_rdf, format="n3")

        pred = util.ns['ir'] + 'isRelevantFor'
        res = {}
        store = TripleStore(self.config.storetype,
                            self.config.storelocation,
                            self.config.storerepository)
        sq_templ = """PREFIX eurlex:<http://lagen.nu/eurlex#>
                      SELECT ?party ?casenum ?celexnum WHERE {
                          <%s> eurlex:party ?party ;
                               eurlex:casenum ?casenum ;
                               eurlex:celexnum ?celexnum .
                      }"""

        self.log.debug(
            "Loading gold standard relevance judgments for %s" % basefile)
        for article in self._articles(basefile):
            res[article] = []
            for o in goldstandard.objects(URIRef(article), URIRef(pred)):
                res[article].append(str(o))
                # Make sure the case exists and is the case we're looking for
                sq = sq_templ % str(o)
                parties = store.select(sq, format="python")
                if parties:
                    pass
                    # self.log.debug("   %s: %s (%s)" %
                    #               (parties[0]['celexnum'],
                    #                parties[0]['casenum'],
                    #                " v ".join([x['party'] for x in parties])))
                else:
                    self.log.warning("Can't find %s in triple store!" % o)
            self.log.debug("    Gold standard for %s: %s relevant docs" %
                           (article, len(res[article])))
            res[article].sort()
        return res

    def eval_aps_chart(self, avg_precisions):
        # Create a chart in PDF format and a equvialent table
        import matplotlib.pyplot as plt
        import numpy as np

        # create linestyle/width array:
        styles = []
        for i in range(1, 5):
            # for j in (['-','--','-.',':']):
            for j in (['-', '-', '-', '-', '-']):
                styles.append((i, j))
        fig = plt.figure()
        fig.set_size_inches(8, 4)
        ax = plt.subplot(111)
        width = len(avg_precisions[0][1])
        plt.axis([0, width - 1, 0, 0.3])
        ind = np.arange(width)
        ax.set_ylabel('Average precision')
        ax.set_title('Average precision for information needs')
        xticklabels = ["Art %d" % int(x.split("E")[-1])
                       for x in self._articles('tfeu')]
        ax.set_xticks(ind)
        ax.set_xticklabels(xticklabels)

        cnt = 0
        for label, aps in avg_precisions:
            # print "%s: %r" % (label, aps)
            lw, ls = styles[cnt]
            plt.plot(ind, aps, label=label, linestyle=ls, linewidth=lw)
            cnt += 1

        plt.legend(loc='best',
                   markerscale=4,
                   prop={'size': 'x-small'},
                   ncol=int(len(avg_precisions) / 8) + 1)

        filetype = self.graph_filetype
        filename = "average_precision"
        filename = self.generic_path(filename, "eval", "." + filetype)

        plt.savefig(filename)
        self.log.info("Created average precision chart as %s" % filename)
        plt.close()
        # plt.show()

    def eval_aps_table(self, avg_precisions):
        filetype = "tex"
        filename = "average_precision"
        filename = self.generic_path(filename, "eval", "." + filetype)

        articles = [x.split("E")[-1] for x in self._articles('tfeu')]
        tblformat = "{r|" + "l|" * len(articles) + "l}"
        tblhead = "".join(["& \\textbf{%s} " % x for x in articles])
        fp = open(filename, "w")
        fp.write("""\\begin{tabular}%s
\\hline
\\textbf{Conf} %s  & \\textbf{MAP}\\\\
\\hline
""" % (tblformat, tblhead))
        for label, aps in avg_precisions:
            if label == "baseline":
                label = "base"
            else:
                label = "".join([x[0].upper() for x in label.split("_")])
            fp.write("%s & " % label)
            for ap in aps:
                fp.write("%.3f & " % ap)
            fp.write("%.3f \\\\ \n" % self.eval_calc_map(aps))
        fp.write("""\\hline
\\end{tabular}
""")
        self.log.info("Created average precision table as %s" % filename)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import math
import re
import shutil
import sys

import six
from six.moves.urllib_parse import quote
import requests
import requests.exceptions
from bs4 import BeautifulSoup

from ferenda import util, errors


class FulltextIndex(object):

    @staticmethod
    def connect(indextype, location, repos=[]):
        """Open a fulltext index (creating it if it
        doesn't already exists).

        :param location: Type of fulltext index ("WHOOSH" or "ELASTICSEARCH")
        :type  location: str
        :param location: The file path of the fulltext index.
        :type  location: str
    """
        # create correct subclass and return it
        return {'WHOOSH': WhooshIndex,
                'ELASTICSEARCH': ElasticSearchIndex}[indextype](location, repos)

    def __init__(self, location, repos):
        self.location = location
        if self.exists():
            self.index = self.open()
        else:
            self.index = self.create(self.get_default_schema(), repos)

    def __del__(self):
        self.close()

    def get_default_schema(self):
        return {'uri': Identifier(),
                'repo': Label(),
                'basefile': Label(),
                'title': Text(boost=4),
                'identifier': Label(boost=16),
                'text': Text()}

    def exists(self):
        """Whether the fulltext index exists."""
        raise NotImplementedError  # pragma: no cover

    def create(self, schema, repos):
        """Creates a fulltext index using the provided default schema."""
        raise NotImplementedError  # pragma: no cover

    def destroy(self):
        """Destroys the index, if created."""
        raise NotImplementedError  # pragma: no cover

    def open(self):
        """Opens the index so that it can be queried."""
        raise NotImplementedError  # pragma: no cover

    def schema(self):
        """Returns the schema that actually is in use. A schema is a dict
           where the keys are field names and the values are any
           subclass of
           :py:class:`ferenda.fulltextindex.IndexedType`
        """
        raise NotImplementedError  # pragma: no cover

    def update(self, uri, repo, basefile, title, identifier, text, **kwargs):
        """Insert (or update) a resource in the fulltext index. A resource may
        be an entire document, but it can also be any part of a
        document that is referenceable (i.e. a document node that has
        ``@typeof`` and ``@about`` attributes). A document with 100
        sections can be stored as 100 independent resources, as long
        as each section has a unique key in the form of a URI.

        :param uri: URI for the resource
        :type  uri: str
        :param repo: The alias for the document repository that the resource is part of
        :type  repo: str
        :param basefile: The basefile which contains resource
        :type  basefile: str
        :param title: User-displayable title of resource (if applicable).
                      Should not contain the same information as
                      ``identifier``. 
        :type  title: str
        :param identifier: User-displayable short identifier for resource (if applicable)
        :type  identifier: str
        :type  text: The full textual content of the resource, as a plain string.
        :type  text: str

        .. note::

           Calling this method may not directly update the fulltext
           index -- you need to call
           :meth:`~ferenda.FulltextIndex.commit` or
           :meth:`~ferenda.FulltextIndex.close` for that.

        """
        raise NotImplementedError  # pragma: no cover

    def commit(self):
        """Commit all pending updates to the fulltext index."""
        raise NotImplementedError  # pragma: no cover

    def close(self):
        """Commits all pending updates and closes the index."""
        raise NotImplementedError  # pragma: no cover

    def doccount(self):
        """Returns the number of currently indexed (non-deleted) documents."""
        raise NotImplementedError  # pragma: no cover

    def query(self, q, **kwargs):
        """Perform a free text query against the full text index, optionally
           restricted with parameters for individual fields.

        :param q: Free text query, using the selected full text index's
                  prefered query syntax
        :type  q: str
        :param \*\*kwargs: any parameter will be used to match a
                         similarly-named field
        :type \*\*kwargs: dict
        :returns: matching documents, each document as a dict of fields
        :rtype: list

        .. note::

           The *kwargs* parameters do not yet do anything -- only
           simple full text queries are possible.

        """
        raise NotImplementedError  # pragma: no cover


class IndexedType(object):

    """Base class for a fulltext searchengine-independent representation
       of indeaxed data.  By using IndexType-derived classes to
       represent the schema, it becomes possible to switch out search
       engines without affecting the rest of the code.

    """

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(tuple(v for k, v in sorted(self.__dict__.items())))

    def __init__(self, **kwargs):
        self.__dict__ = dict(kwargs)

    def __repr__(self):
        # eg '<Label boost=16>' or '<Identifier>'
        dictrepr = "".join((" %s=%s" % (k, v) for k, v in sorted(self.__dict__.items())))
        return ("<%s%s>" % (self.__class__.__name__, dictrepr))


class Identifier(IndexedType):
    pass


class Datetime(IndexedType):
    pass


class Text(IndexedType):
    pass


class Label(IndexedType):
    pass


class Keywords(IndexedType):
    pass


class Boolean(IndexedType):
    pass


class URI(IndexedType):
    pass


class SearchModifier(object):
    pass


class Less(SearchModifier):
    pass


class More(SearchModifier):
    pass


class Between(SearchModifier):
    pass


import whoosh.index
import whoosh.fields
import whoosh.analysis
import whoosh.query
import whoosh.qparser
import whoosh.writing
import whoosh.highlight

from ferenda.elements import html


class ElementsFormatter(whoosh.highlight.Formatter):

    """Returns a tree of ferenda.elements representing the formatted hit."""

    def __init__(self, wrapelement=html.P, hitelement=html.Strong, classname="match", between=" ... "):
        self.wrapelement = wrapelement
        self.hitelement = hitelement
        self.classname = classname
        self.between = between

    def format(self, fragments, replace=False):
        res = self.wrapelement()
        first = True
        for fragment in fragments:
            if not first:
                res.append(self.between)
            res.extend(self.format_fragment(fragment, replace=replace))
            first = False
        return res

    re_collapse = re.compile("\s+").sub

    def format_fragment(self, fragment, replace):
        output = []
        index = fragment.startchar
        text = fragment.text

        for t in fragment.matches:
            if t.startchar > index:
                output.append(self.re_collapse(" ", text[index:t.startchar]))
            hittext = whoosh.highlight.get_text(text, t, False)
            output.append(self.hitelement([hittext], **{'class': self.classname}))
            index = t.endchar
        if index < len(text):
            output.append(self.re_collapse(" ", text[index:fragment.endchar]))
        return output


class WhooshIndex(FulltextIndex):

    def __init__(self, location, repos):
        super(WhooshIndex, self).__init__(location, repos)
        self._schema = self.get_default_schema()
        self._writer = None

    def exists(self):
        return whoosh.index.exists_in(self.location)

    def open(self):
        return whoosh.index.open_dir(self.location)

    def create(self, schema, repos):
        # maps our field classes to concrete whoosh field instances
        mapped_field = {Identifier():   whoosh.fields.ID(unique=True, stored=True),
                        Label():        whoosh.fields.ID(stored=True),
                        Label(boost=16): whoosh.fields.ID(field_boost=16, stored=True),
                        Text(boost=4):  whoosh.fields.TEXT(field_boost=4, stored=True,
                                                           analyzer=whoosh.analysis.StemmingAnalyzer(
                                                           )),
                        Text():         whoosh.fields.TEXT(stored=True,
                                                           analyzer=whoosh.analysis.StemmingAnalyzer())}

        whoosh_fields = {}
        for key, fieldtype in self.get_default_schema().items():
            whoosh_fields[key] = mapped_field[fieldtype]
        schema = whoosh.fields.Schema(**whoosh_fields)
        util.mkdir(self.location)
        return whoosh.index.create_in(self.location, schema)

    def destroy(self):
        shutil.rmtree(self.location)

    def schema(self):
        # FIXME: This should iterate through self.index (the
        # underlying whoosh index), convert each field to the
        # corresponding IndexedType objects.
        return self._schema

    def update(self, uri, repo, basefile, title, identifier, text, **kwargs):
        if not self._writer:
            self._writer = self.index.writer()

        # A whoosh document is not the same as a ferenda document. A
        # ferenda document may be indexed as several (tens, hundreds
        # or more) whoosh documents
        self._writer.update_document(uri=uri,
                                     repo=repo,
                                     basefile=basefile,
                                     title=title,
                                     identifier=identifier,
                                     text=text,
                                     **kwargs)

    def commit(self):
        if self._writer:
            self._writer.commit()
            if not isinstance(self._writer, whoosh.writing.BufferedWriter):
                # A bufferedWriter can be used again after commit(), a regular writer cannot
                self._writer = None

    def close(self):
        self.commit()
        self.index.close()

    def doccount(self):
        return self.index.doc_count()

    def query(self, q, pagenum=1, pagelen=10, **kwargs):
        searchfields = ['identifier', 'title', 'text']
        mparser = whoosh.qparser.MultifieldParser(searchfields,
                                                  self.index.schema)
        query = mparser.parse(q)
        with self.index.searcher() as searcher:
            page = searcher.search_page(query, pagenum, pagelen)
            res = self._convert_result(page)
            pager = {'pagenum': pagenum,
                     'pagecount': page.pagecount,
                     'firstresult': page.offset + 1,
                     'lastresult': page.offset + page.pagelen,
                     'totalresults': page.total}
        return res, pager

    def _convert_result(self, res):
        # converts a whoosh.searching.ResultsPage object to a plain
        # list of dicts
        l = []
        hl = whoosh.highlight.Highlighter(formatter=ElementsFormatter())
        for hit in res:
            fields = hit.fields()
            fields['text'] = hl.highlight_hit(hit, "text", fields['text'])
            l.append(hit.fields())
        return l

# Base class for a HTTP-based API (eg. ElasticSearch) the base class
# delegate the formulation of queries, updates etc to concrete
# subclasses, expected to return a formattted query/payload etc, and
# be able to decode responses to queries, but the base class handles
# the actual HTTP call, inc error handling.


class RemoteIndex(FulltextIndex):

    # The only real implementation of RemoteIndex has its own exists
    # implementation, no need for a general fallback impl.
    # def exists(self):
    #     pass

    def create(self, schema, repos):
        relurl, payload = self._create_schema_payload(self.get_default_schema(), repos)
        res = requests.put(self.location + relurl, payload)
        try:
            res.raise_for_status()
        except Exception as e:
            raise Exception("%s: %s" % (res.status_code, res.text))

    def schema(self):
        relurl, payload = self._get_schema_payload()
        res = requests.get(self.location + relurl)  # payload is probably never used
        return self._decode_schema(res)

    def update(self, uri, repo, basefile, title, identifier, text, **kwargs):
        relurl, payload = self._update_payload(
            uri, repo, basefile, title, identifier, text, **kwargs)

        res = requests.put(self.location + relurl, payload)
        try:
            res.raise_for_status()
            # print(json.dumps(res.json(), indent=4))
        except requests.exceptions.HTTPError as e:
            raise errors.IndexingError(str(e) + ": '%s'" % res.text)

    def doccount(self):
        relurl, payload = self._count_payload()
        if payload:
            res = requsts.post(self.location + relurl, payload)
        else:
            res = requests.get(self.location + relurl)
        return self._decode_count_result(res)

    def query(self, q, pagenum=1, pagelen=10, **kwargs):
        relurl, payload = self._query_payload(q, pagenum, pagelen, **kwargs)
        if payload:
            # print("POSTing to %s:\n%s" % (relurl, payload))
            res = requests.post(self.location + relurl, payload)
            # print("Recieved:\n%s" % (json.dumps(res.json(),indent=4)))
        else:
            res = requests.get(self.location + relurl)
        try:
            res.raise_for_status()
        except Exception as e:
            raise errors.SearchingError("%s: %s" % (res.status_code, res.text))
        return self._decode_query_result(res, pagenum, pagelen)

    def destroy(self):
        reluri, payload = self._destroy_payload()
        res = requests.delete(self.location + reluri)

    # these don't make no sense for a remote index accessed via HTTP/REST
    def open(self):
        pass

    def commit(self):
        pass

    def close(self):
        pass


class ElasticSearchIndex(RemoteIndex):

    def commit(self):
        r = requests.post(self.location + "_refresh")
        r.raise_for_status()

    def exists(self):
        r = requests.get(self.location + "_mapping/")
        if r.status_code == 404:
            return False
        else:
            return True

    def _update_payload(self, uri, repo, basefile, title, identifier, text, **kwargs):
        safe = ''
        if six.PY2:
            # urllib.quote in python 2.6 cannot handle unicode values
            # for the safe parameter (not even empty). urllib.quote in
            # python 2.7 handles it, but may fail later on. FIXME: We
            # should create a shim as ferenda.compat.quote and use
            # that
            safe = safe.encode('ascii') # pragma: no cover

        # quote (in python 2) only handles characters from 0x0 - 0xFF,
        # and basefile might contain characters outside of that (eg
        # u'MO\u0308D/P11463-12', which is MÖD/P11463-12 on a system
        # which uses unicode normalization form NFD). To be safe,
        # encodethe string to utf-8 beforehand (Which is what quote on
        # python 3 does anyways)
        relurl = "%s/%s" % (repo, quote(basefile.encode("utf-8"), safe=safe))  # eg type, id
        if "#" in uri:
            relurl += uri.split("#", 1)[1]
        payload = {"uri": uri,
                   "basefile": basefile,
                   "title": title,
                   "identifier": identifier,
                   "text": text}
        payload.update(kwargs)
        return relurl, json.dumps(payload)

    def _query_payload(self, q, pagenum=1, pagelen=10, **kwargs):
        # relurl = "_search?q=%s&size=%s&from=%s" % (quote(q), pagelen, (pagenum * pagelen) - pagelen)
        relurl = "_search?from=%s&size=%s" % ((pagenum - 1) * pagelen, pagelen)

        # FIXME: Only searches in text, not title or identifier. But
        # can't search on the _all field, because apparently that
        # field isn't set up to use the my_analyzer we've defined...
        payload = {'query': {'match': {'text': q}},
                   'highlight': {'fields': {'text': {}},
                                 'pre_tags': ["<strong class='match'>"],
                                 'post_tags': ["</strong>"],
                                 'fragment_size': '40'}}
        return relurl, json.dumps(payload, indent=4)

    def _decode_query_result(self, response, pagenum, pagelen):
        json = response.json()
        res = []
        for hit in json['hits']['hits']:
            h = hit['_source']
            # wrap highlighted field in P, convert to elements
            hltext = " ... ".join([x.strip() for x in hit['highlight']['text']])
            soup = BeautifulSoup("<p>%s</p>" % re.sub("\s+", " ", hltext))
            h['text'] = html.elements_from_soup(soup.html.body.p)
            res.append(h)
        pager = {'pagenum': pagenum,
                 'pagecount': int(math.ceil(json['hits']['total'] / float(pagelen))),
                 'firstresult': (pagenum - 1) * pagelen + 1,
                 'lastresult': (pagenum - 1) * pagelen + len(json['hits']['hits']),
                 'totalresults': json['hits']['total']}
        return res, pager

    def _count_payload(self):
        return "_count", None

    def _decode_count_result(self, response):
        if response.status_code == 404:
            return 0
        else:
            return response.json()['count']

    # FIXME: This is cheating!
    def schema(self):
        return self.get_default_schema()

    def _get_schema_payload(self):
        return "", None

    def _decode_schema_payload(self, response):
        raise NotImplementedError # pragma: no cover

    # FIXME: For some reason, createing a schema/mapping makes PUTting
    # new documents to the index hang with the folloging error:
    #
    #    UnavailableShardsException[[ferenda][1] [3] shardIt, [0] active : Timeout waiting for [1m]
    #
    # So we skip creating the schema as it isn't neccesary
    # def create(self, schema, repos):
    #    pass

    def _create_schema_payload(self, schema, repos):
        schema = {
            # cargo cult configuration
            "settings": {"number_of_shards": 1,
                         "analysis": {
                             "analyzer": {
                                 "my_analyzer": {
                                     "filter": ["lowercase", "snowball"],
                                     "tokenizer": "standard",
                                     "type": "custom"
                                 }
                             },
                             "filter": {
                                 "snowball": {
                                     "type": "snowball",
                                     "language": "English"
                                 }
                             }
                         }
                         },
            # "mappings": {"_all": {"properties": {"analyzer": "my_analyzer"}}}
            "mappings": {}
        }

        # maps our field classes to concrete ES field properties
        mapped_field = {Identifier():   {"type": "string", "index": "not_analyzed"},  # uri
                        # repo, basefile (note: see below)
                        Label():        {"type": "string", "index": "not_analyzed"},
                        # identifier
                        Label(boost=16): {"type": "string", "boost": 16.0, "analyzer": "my_analyzer"},
                        # title
                        Text(boost=4):  {"type": "string", "boost": 4.0, "analyzer": "my_analyzer"},
                        Text():         {"type": "string", "analyzer": "my_analyzer"}}  # text

        es_fields = {}
        for key, fieldtype in self.get_default_schema().items():
            if key == "repo":
                continue  # not really needed for ES, as type == repo.alias
            es_fields[key] = mapped_field[fieldtype]
        for repo in repos:
            schema["mappings"][repo.alias] = {"_source": {"enabled": True},  # so we can get the text back
                                              "properties": es_fields}
        return "", json.dumps(schema)

    def _destroy_payload(self):

        return "", None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os,sys
import codecs
from ferenda.compat import unittest
if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())

import six

from ferenda import CitationParser
from ferenda import util
import ferenda.citationpatterns
from ferenda.testutil import file_parametrize

class ParametricBase(unittest.TestCase):
    parser = ferenda.citationpatterns.url
    def parametric_test(self,filename):
        with codecs.open(filename,encoding="utf-8") as fp:
            testdata = fp.read()
        
        cp = CitationParser(self.parser)
        nodes = cp.parse_string(testdata)
        got = []
        for node in nodes:
            if isinstance(node,six.text_type):
                got.append(node.strip())
            else:
                (text,result) = node
                got.append(util.parseresults_as_xml(result).strip())
        
        wantfile = os.path.splitext(filename)[0] + ".result"
        if os.path.exists(wantfile):
            with open(wantfile) as fp:
                want = [x.strip() for x in fp.read().split("\n\n")]
        else:
            print("\nparse_string() returns:")
            print("\n\n".join(compare))
            self.fail("%s not found" % wantfile)
        self.maxDiff = 4096
        self.assertListEqual(want,got)

class URL(ParametricBase):
    parser = ferenda.citationpatterns.url

class EULaw(ParametricBase):
    parser = ferenda.citationpatterns.eulaw

file_parametrize(URL, "test/files/citation/url", ".txt")
# file_parametrize(URL, "test/files/citation/eulaw", ".txt")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os, tempfile, shutil
from datetime import date
from six import text_type as str
from lxml import etree
from bs4 import BeautifulSoup
from rdflib import Graph
from ferenda.compat import unittest

if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())
from ferenda.manager import setup_logger; setup_logger('CRITICAL')
from ferenda.citationpatterns import url as urlparser
from ferenda import util
# SUT
from ferenda.elements import serialize, deserialize, AbstractElement, UnicodeElement, CompoundElement, TemporalElement, OrdinalElement, PredicateElement, Body, Section, Paragraph, Link, html

class Main(unittest.TestCase):

    def test_serialize_roundtrip(self):

        # Create a elements object tree
        tree = Body([Section([Paragraph(["Hello"]),
                              Paragraph(["World"])],
                             ordinal="1",
                             title="Main section"),
                     Section([42,
                              date(2013,11,27),
                              b'bytestring',
                              {'foo': 'bar',
                               'x': 'y'}],
                             ordinal=2,
                             title="Native types")
                 ])
        # roundtrip using the default XML format 
        serialized = serialize(tree)
        self.assertIsInstance(serialized, str)
        newtree = deserialize(serialized, caller_globals=globals())
        self.assertEqual(tree, newtree)

        # make another section with special (but commonly used) types
        # and try to roundtrip them. The XML serialization format does
        # not support this.
        graph = Graph().parse(data="""@prefix dct: <http://purl.org/dc/terms/> .

<http://example.org/1> dct:title "Hello world"@en .
""", format="turtle")
        parseresult = urlparser.parseString("http://example.org/1")
        tree.append(Section([parseresult,
                             graph],
                            meta=graph))
        
        # roundtrip using JSON (which uses fully qualified classnames,
        # so we don't need to pass globals() into deserialize()
        serialized = serialize(tree, format="json")
        self.assertIsInstance(serialized, str)
        newtree = deserialize(serialized, format="json")

        # two pyparsing.ParseResult objects cannot be directly
        # compared (they don't implement __eq__), therefore we compare
        # their XML representations
        tree[2][0] = util.parseresults_as_xml(tree[2][0])
        newtree[2][0] = util.parseresults_as_xml(newtree[2][0])
        self.assertEqual(tree, newtree)

    def test_json_roundtrip(self):
        # a more realistic roundtrip example with some hairy parts
        from ferenda import PDFDocumentRepository, PDFReader
        d = PDFDocumentRepository()
        doc = d.make_document("sample")
        reader = PDFReader()
        reader.read("test/files/pdfreader/sample.pdf",
                    "test/files/pdfreader/intermediate")
        d.parse_from_pdfreader(reader, doc)
        jsondoc = serialize(doc, format="json")
        newdoc = deserialize(jsondoc, format="json")
        self.assertEqual(doc, newdoc)

    def test_serialize_pyparsing(self):
        # these objects can't be roundtripped
        from ferenda.citationpatterns import url
        x = url.parseString("http://example.org/foo?param=val")
        serialized = serialize(Body([x]))
        self.assertEqual("""<Body>
  <url>
    <scheme>http</scheme>
    <netloc>example.org</netloc>
    <path>/foo</path>
    <query>param=val</query>
  </url>
</Body>
""", serialized)
        
        

    def test_abstract(self):
        x = AbstractElement()
        with self.assertRaises(AttributeError):
            x.foo = "bar"

        self.assertEqual(b'<abstractelement xmlns="http://www.w3.org/1999/xhtml"/>',
                         etree.tostring(x.as_xhtml()))
        

    def test_compound(self):
        x = CompoundElement(["hello", "world"], id="42", foo="bar")
        x.foo = "baz"
        with self.assertRaises(AttributeError):
            x.y = "z"
        x.append(os.listdir) # a non-serializable object (in this case a function)
        self.assertEqual(b'<compoundelement xmlns="http://www.w3.org/1999/xhtml" id="42">helloworld&lt;built-in function listdir&gt;</compoundelement>',
                         etree.tostring(x.as_xhtml()))
        self.assertEqual(Body([Section([Paragraph(["Hello"]),
                                        Paragraph(["World"])])]).as_plaintext(),
                         "Hello World")
        

    def test_unicode(self):
        x = UnicodeElement("Hello world", id="42")
        self.assertEqual(b'<unicodeelement xmlns="http://www.w3.org/1999/xhtml" id="42">Hello world</unicodeelement>',
                         etree.tostring(x.as_xhtml()))

        with self.assertRaises(TypeError):
            UnicodeElement(b'bytestring')
        
    def test_temporal(self):
        class TemporalString(UnicodeElement, TemporalElement): pass
        x = TemporalString("Hello", entryintoforce=date(2013,1,1),
                           expires=date(2014,1,1))
        self.assertFalse(x.in_effect(date(2012,7,1)))
        self.assertTrue(x.in_effect(date(2013,7,1)))
        self.assertFalse(x.in_effect(date(2014,7,1)))
                                                        
    def test_ordinal(self):
        class OrdinalString(UnicodeElement, OrdinalElement): pass
        x = OrdinalString("Foo", ordinal="2")
        y = OrdinalString("Bar", ordinal="2 a")
        z = OrdinalString("Baz", ordinal="10")
        w = OrdinalString("Duplicate of Foo", ordinal="2")
        self.assertTrue(x < y < z)
        self.assertTrue(z > y > x)
        self.assertTrue(x != y)
        self.assertTrue(x == w)
        self.assertTrue(x <= w <= y)
        self.assertTrue(y >= w >= x)
        
    def test_predicate(self):
        class PredicateString(UnicodeElement, PredicateElement): pass
        # known vocabulary used
        x = PredicateString("This is my title", predicate="http://purl.org/dc/terms/title")
        self.assertEqual("dct:title", x.predicate)

        # unknown vocabulary used
        y = PredicateString("This is my title", predicate="http://example.org/vocab/title")
        self.assertEqual("http://example.org/vocab/title", y.predicate)

        # No predicate used --- default to rdfs:Resource
        z = PredicateString("This is a resource")
        from rdflib import RDFS
        self.assertEqual(RDFS.Resource, z.predicate)

    def test_link(self):
        x = Link("Link text", uri="http://example.org/")
        self.assertEqual(str(x), "Link text")
        self.assertEqual(repr(x), "Link('Link text', uri=http://example.org/)")

        # y = Link("Räksmörgås", uri="http://example.org/")
        # self.assertEqual(str(y), "Räksmörgås")
        # self.assertEqual(repr(y), "Link('Räksmörgås', uri=http://example.org/)")
        
    def test_elements_from_soup(self):
        soup = BeautifulSoup("""<html>
<head>
  <title>Example doc</title>
</head>
<body>
  <marquee>Hello world</marquee>
  <!-- Hello world -->
  <center>Hello world</center>
  <p>That's enough of this nonsense</p>
</body>""")
        got = html.elements_from_soup(soup.html)
        self.assertEqual(html.HTML([html.Head([html.Title(["Example doc"])]),
                                    html.Body([html.P(["That's enough of this nonsense"])])]),
                         got)

        
import doctest
def load_tests(loader,tests,ignore):
    from ferenda.elements import elements
    tests.addTests(doctest.DocTestSuite(elements))
    return tests



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# A abstract base class for fetching documents from data.riksdagen.se

import os
import codecs

import requests
import requests.exceptions
from bs4 import BeautifulSoup


from . import SwedishLegalSource
from .swedishlegalsource import offtryck_parser, offtryck_gluefunc, PreambleSection, UnorderedSection
from ferenda import util, errors
from ferenda.decorators import managedparsing, downloadmax
from ferenda.describer import Describer
from ferenda.elements import Paragraph
from ferenda import PDFReader

class Riksdagen(SwedishLegalSource):
    BILAGA = "bilaga"
    DS = "ds"
    DIREKTIV = "dir"
    EUNAMND_KALLELSE = "kf-lista"
    EUNAMND_PROT = "eunprot"
    EUNAMND_DOK = "eundok"
    EUNAMND_BILAGA = "eunbil"
    FAKTAPROMEMORIA = "fpm"
    FRAMSTALLNING = "frsrdg"
    FOREDRAGNINGSLISTA = "f-lista"
    GRANSKNINGSRAPPORT = "rir"
    INTERPELLATION = "ip"
    KOMMITTEBERATTELSER = "komm"
    MINISTERRADSPROMEMORIA = "minråd"
    MOTION = "mot"
    PROPOSITION = "prop"
    PROTOKOLL = "prot"
    RAPPORT = "rfr"
    RIKSDAGSSKRIVELSE = "rskr"
    FRAGA = "fr"
    SKRIVELSE = "skr"
    SOU = "sou"
    SVAR = "frs"
    SFS = "sfs"
    TALARLISTA = "t-lista"
    UTSKOTTSDOKUMENT = "utskottsdokument"
    YTTRANDE = "yttr"

    # add typ=prop or whatever
    downloaded_suffix = ".xml"
    storage_policy = "dir"

    document_type = None
    start_url = None
    start_url_template = "http://data.riksdagen.se/dokumentlista/?sz=100&sort=d&utformat=xml&typ=%(doctype)s"

    def download(self, basefile=None):
        if basefile:
            return self.download_single(basefile)
        url = self.start_url_template % {'doctype': self.document_type}
        for basefile, url in self.download_get_basefiles(url):
            self.download_single(basefile, url)

    @downloadmax
    def download_get_basefiles(self, start_url):
        self.log.info("Starting at %s" % start_url)
        url = start_url
        done = False
        pagecount = 1
        while not done:
            resp = requests.get(url)
            soup = BeautifulSoup(resp.text, features="xml")

            subnodes = soup.find_all(lambda tag: tag.name == "subtyp" and
                                     tag.text == self.document_type)
            for doc in [x.parent for x in subnodes]:
                # TMP: Only retrieve old documents
                # if doc.rm.text > "1999":
                #     continue
                if doc.rm is None or doc.nummer is None:
                    # this occasionally happens, although not all the
                    # time for the same document. We can't really
                    # recover from this easily, so we'll just skip and
                    # hope that it doesn't occur on next download.
                    continue
                basefile = "%s:%s" % (doc.rm.text, doc.nummer.text)
                attachment = None
                if doc.tempbeteckning.text:
                    attachment = doc.tempbeteckning.text
                yield (basefile, attachment), doc.dokumentstatus_url_xml.text
            try:
                url = soup.dokumentlista['nasta_sida']
                pagecount += 1
                self.log.info("Getting page #%d" % pagecount)
            except KeyError:
                self.log.info("That was the last page")
                done = True

    def download_single(self, basefile, url=None):
        if isinstance(basefile, tuple):
            basefile, attachment = basefile
        if attachment:
            docname = attachment
        else:
            docname = "index"

        if not url:
            url = self.remote_url(basefile)
            if not url:  # remote_url failed
                return False

        xmlfile = self.store.downloaded_path(basefile)
        if not (self.config.force or not os.path.exists(xmlfile)):
            self.log.debug("%s already exists" % (xmlfile))
            return False

        existed = os.path.exists(xmlfile)
        self.log.debug("  %s: Downloading to %s" % (basefile, xmlfile))

        updated = self.download_if_needed(url, basefile)

        if existed:
            if updated:
                self.log.info("%s: updated from %s" % (basefile, url))
            else:
                self.log.debug("%s: %s is unchanged, checking files" %
                               (basefile, xmlfile))
        else:
            self.log.info("%s: downloaded from %s" % (basefile, url))
        fileupdated = False
        r = None
        docsoup = BeautifulSoup(open(xmlfile), features="xml")
        dokid = docsoup.find('dok_id').text
        if docsoup.find('dokument_url_html'):
            htmlurl = docsoup.find('dokument_url_html').text
            htmlfile = self.store.downloaded_path(basefile, attachment=docname + ".html")
            self.log.debug("   Downloading to %s" % htmlfile)
            r = self.download_if_needed(htmlurl, basefile, filename=htmlfile)
        elif docsoup.find('dokument_url_text'):
            texturl = docsoup.find('dokument_url_text').text
            textfile = self.store.downloaded_path(basefile, attachment=docname + ".txt")
            self.log.debug("   Downloading to %s" % htmlfile)
            r = self.download_if_needed(texturl, basefile, filename=textfile)
        fileupdated = fileupdated or r

        for b in docsoup.findAll('bilaga'):
            # self.log.debug("Looking for %s, found %s", dokid, b.dok_id.text)
            if b.dok_id.text != dokid:
                continue
            if b.filtyp is None:
                # apparantly this can happen sometimes? Very intermitently, though.
                self.log.warning("Couldn't find filtyp for bilaga %s in %s" % (b.dok_id.text, xmlfile))
                continue
            filetype = "." + b.filtyp.text
            filename = self.store.downloaded_path(basefile, attachment=docname + filetype)
            self.log.debug("   Downloading to %s" % filename)
            try:
                r = self.download_if_needed(b.fil_url.text, basefile, filename=filename)
            except requests.exceptions.HTTPError as e:
                # occasionally we get a 404 even though we shouldn't. Report and hope it goes better next time.
                self.log.error("   Failed: %s" % e)
                continue
            fileupdated = fileupdated or r
            break

        if updated or fileupdated:
            return True  # Successful download of new or changed file
        else:
            self.log.debug(
                "%s and all associated files unchanged" % xmlfile)

    @managedparsing
    def parse(self, doc):
        filename = self.store.downloaded_path(doc.basefile)
        if not os.path.exists(filename):
            raise errors.NoDownloadedFileError("File '%s' not found" % filename)
        htmlfile = self.store.path(doc.basefile, 'downloaded', '.html')
        pdffile  = self.store.path(doc.basefile, 'downloaded', '.pdf')

        intermediate_path = self.store.intermediate_path(doc.basefile,
                                                         attachment=os.path.basename(pdffile))
        intermediate_dir = os.path.dirname(intermediate_path)

        if os.path.exists(pdffile):
            parser = offtryck_parser(preset='proposition')
            pdf = PDFReader()
            pdf.read(pdffile, workdir=intermediate_dir, images=False, keep_xml="bz2")
            if pdf.is_empty():
                self.log.debug("%s: %s contains no text, performing OCR" % (doc.basefile, pdffile))
                pdf.read(pdffile, workdir=intermediate_dir, ocr_lang="swe", keep_xml="bz2")
            doc.body = parser.parse(pdf.textboxes(offtryck_gluefunc))
        else:
            self.log.debug("Loading soup from %s" % htmlfile)
            soup = BeautifulSoup(
                codecs.open(
                    htmlfile, encoding='iso-8859-1', errors='replace').read(),
                )
            self.parse_from_soup(soup, doc)
        doc.uri = self.canonical_uri(doc.basefile)
        self.log.debug("Set URI to %s (from %s)" % (doc.uri, doc.basefile))
        d = Describer(doc.meta, doc.uri)
        d.rdftype(self.rdf_type)
        d.value(self.ns['prov'].wasGeneratedBy, self.qualified_class_name())
        xsoup = BeautifulSoup(open(self.store.downloaded_path(doc.basefile)).read(), "xml")
        d.value(self.ns['dct'].title, xsoup.dokument.titel.text, lang="sv")
        d.value(self.ns['dct'].issued, util.strptime(xsoup.dokument.publicerad.text,
                                                        "%Y-%m-%d %H:%M:%S").date())
        self.infer_triples(d, doc.basefile)
        return True

    def parse_from_soup(self, soup, doc):
        for block in soup.findAll(['div', 'p', 'span']):
            t = util.normalize_space(''.join(block.findAll(text=True)))
            block.extract()  # to avoid seeing it again
            if t:
                doc.body.append(Paragraph([t]))

    def canonical_uri(self, basefile):
        seg = {self.ns['rpubl'].Proposition: "prop",
               self.ns['rpubl'].Skrivelse: "skr"}
        return self.config.url + "res/%s/%s" % (seg[self.rdf_type], basefile)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rdflib import Graph


class Document(object):

    """A document represents the content of a document together with a
    RDF graph containing metadata about the document. Don't create
    instances of :class:`~ferenda.Document` directly. Create them
    through :meth:`~ferenda.DocumentRepository.make_document` in order
    to properly initialize the ``meta`` property.

    :param meta: A RDF graph containing metadata about the document
    :param body: A list of :mod:`ferenda.elements` based objects representing the content of the document
    :param uri: The canonical URI for this document
    :param lang: The main language of the document as a IETF language tag, i.e. "sv" or "en-GB"
    :param basefile: The basefile of the document
    """

    def __init__(self, meta=None, body=None, uri=None, lang=None, basefile=None):
        if meta is None:
            self.meta = Graph()
        else:
            self.meta = meta
        if body is None:
            self.body = []
        else:
            self.body = body
        self.uri = uri
        self.lang = lang
        self.basefile = basefile

    def __eq__(self, other):
        return self.__dict__ == other.__dict__



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import re
import os
from datetime import datetime
import codecs

from bs4 import BeautifulSoup
from lxml import etree
import requests
from six import text_type as str

from ferenda import util, errors
from ferenda.elements import UnicodeElement, CompoundElement, \
    Heading, Preformatted, Paragraph, Section, Link, ListItem, \
    serialize
from ferenda import CompositeRepository, CompositeStore
from ferenda import PDFDocumentRepository
from ferenda import Describer
from ferenda import TextReader
from ferenda import PDFReader
from ferenda import DocumentEntry
from ferenda import LayeredConfig
from ferenda.decorators import managedparsing
from . import Trips, NoMoreLinks
from . import Regeringen
from . import Riksdagen
from . import RPUBL
from . import SwedishLegalSource, SwedishLegalStore
from .swedishlegalsource import offtryck_parser, offtryck_gluefunc

class PropPolo(Regeringen):
    alias = "proppolo"
    re_basefile_strict = re.compile(r'Prop. (\d{4}/\d{2,4}:\d+)')
    re_basefile_lax = re.compile(
        r'(?:Prop\.?|) ?(\d{4}/\d{2,4}:\d+)', re.IGNORECASE)
    rdf_type = RPUBL.Proposition
    document_type = Regeringen.PROPOSITION


class PropTrips(Trips):
    alias = "proptrips"
    base = "THWALLAPROP"
    app = "prop"

    basefile_regex = "(?P<basefile>\d+/\d+:\d+)$"
    download_params = [{'maxpage': 101, 'app': app, 'base': base}]

    downloaded_suffix = ".html"
    rdf_type = RPUBL.Proposition

    storage_policy = "dir"

    def get_default_options(self):
        opts = super(PropTrips, self).get_default_options()
        opts['lastbase'] = "THWALLAPROP"
        return opts

    # don't use @recordlastdownload -- download_get_basefiles_page
    # should set self.config.lastbase instead
    def download(self, basefile=None):
        if basefile:
            return super(PropTrips, self).download(basefile)
        else:
            if (hasattr(self.config, 'lastbase') and
                self.config.lastbase and
                not self.config.refresh):
                now = datetime.now()
                maxbase = "PROPARKIV%s%s" % (now.year % 100, (now.year+1) % 100)
                while self.config.lastbase != maxbase:
                    self.download_params[0]['base'] = self.config.lastbase  # override "THWALLAPROP" with eg "PROPARKIV0809"
                    r = super(PropTrips, self).download() # download_get_basefiles_page sets lastbase as it goes along
            else:
                r = super(PropTrips, self).download()
            LayeredConfig.write(self.config)      # assume we have data to write
            return r


    def download_get_basefiles_page(self, pagetree):
        def basefile_to_base(basefile):
            # 1992/93:23 -> "PROPARKIV9293"
            # 1999/2000:23 -> "PROPARKIV9900"
            (y1, y2, idx) = re.split("[:/]", basefile)
            return "PROPARKIV%02d%02d" % (int(y1) % 100, int(y2) % 100)

        # feed the lxml tree into beautifulsoup by serializing it to a
        # string -- is there a better way?
        soup = BeautifulSoup(etree.tostring(pagetree))
        for tr in soup.findAll("tr"):
            if ((not tr.find("a")) or
                    not re.match(self.basefile_regex, tr.find("a").text)):
                # FIXME: Maybe re.search instead of .match to find
                # "Prop. 2012/13:152"
                continue
            # First, look at desc (third td):
            descnodes = [util.normalize_space(x) for x
                         in tr.find_all("td")[2]
                         if isinstance(x, str)]
            bilaga = None
            if len(descnodes) > 1:
                if descnodes[1].startswith("Bilaga:"):
                    bilaga = util.normalize_space(descnodes[0].split(",")[-1])
            desc = "\n".join(descnodes)

            # then, find basefile (second td)
            tds = tr.find_all("td")
            td = tds[1]
            basefile = td.a.text
            assert re.match(self.basefile_regex, basefile)

            basefile = self.sanitize_basefile(basefile)

            # assume entries are strictly sorted from ancient to
            # recent. Therefore, as soon as we encounter a new time
            # period (eg 1998/99) we can update self.config.lastbase
            self.config.lastbase = basefile_to_base(basefile)

            url = td.a['href']

            # self.download_single(basefile, refresh=refresh, url=url)

            # and, if present, extra files (in td 4+5)
            extraurls = []
            for td in tr.findAll("td")[3:]:
                extraurls.append(td.a['href'])

            # we slightly abuse the protocol between
            # download_get_basefiles and this generator -- instead of
            # yielding just two strings, we yield two tuples with some
            # extra information that download_single will need.

            yield (basefile, bilaga), (url, extraurls)

        nextpage = None
        for element, attribute, link, pos in pagetree.iterlinks():
            if element.text and element.text.strip() == "Fler poster":
                nextpage = link
        raise NoMoreLinks(nextpage)

    def download_single(self, basefile, url):
        # unpack the tuples we may recieve instead of plain strings
        if isinstance(basefile, tuple):
            basefile, attachment = basefile
            if attachment:
                mainattachment = attachment + ".html"
            else:
                mainattachment = None
        if isinstance(url, tuple):
            url, extraurls = url
        updated = created = False
        checked = True

        filename = self.store.downloaded_path(basefile, attachment=mainattachment)
        created = not os.path.exists(filename)

        # since the server doesn't support conditional caching and
        # propositioner are basically never updated once published, we
        # avoid even calling download_if_needed if we already have
        # the doc.
        if (os.path.exists(self.store.downloaded_path(basefile))
            and not self.config.refresh):
            self.log.debug("%s: already exists" % basefile)
            return
        
        if self.download_if_needed(url, basefile, filename=filename):
            if created:
                self.log.info("%s: downloaded from %s" % (basefile, url))
            else:
                self.log.info(
                    "%s: downloaded new version from %s" % (basefile, url))
            updated = True
        else:
            self.log.debug("%s: exists and is unchanged" % basefile)

        for url in extraurls:
            if url.endswith('msword.application'):
                # NOTE: We cannot be sure that this is
                # actually a Word (CDF) file. For older files
                # it might be a WordPerfect file (.wpd) or a
                # RDF file, for newer it might be a .docx. We
                # cannot be sure until we've downloaded it.
                # So we quickly read the first 4 bytes
                r = requests.get(url, stream=True)
                sig = r.raw.read(4)
                # r.raw.close()
                #bodyidx = head.index("\n\n")
                #sig = head[bodyidx:bodyidx+4]
                if sig == b'\xffWPC':
                    doctype = ".wpd"
                elif sig == b'\xd0\xcf\x11\xe0':
                    doctype = ".doc"
                elif sig == b'PK\x03\x04':
                    doctype = ".docx"
                elif sig == b'{\\rt':
                    doctype = ".rtf"
                else:
                    self.log.error(
                        "%s: Attached file has signature %r -- don't know what type this is" % (basefile, sig))
                    continue
            elif url.endswith('pdf.application'):
                doctype = ".pdf"
            else:
                self.log.warning("Unknown doc type %s" %
                                 td.a['href'].split("=")[-1])
                doctype = None
            if doctype:
                if attachment:
                    filename = self.store.downloaded_path(
                        basefile, attachment=attachment + doctype)
                else:
                    filename = self.store.downloaded_path(basefile, attachment="index" + doctype)
                self.log.debug("%s: downloading attachment %s" % (basefile, filename))
                self.download_if_needed(url, basefile, filename=filename)

        if mainattachment == None:
            entry = DocumentEntry(self.store.documententry_path(basefile))
            now = datetime.now()
            entry.orig_url = url
            if created:
                entry.orig_created = now
            if updated:
                entry.orig_updated = now
            if checked:
                entry.orig_checked = now
            entry.save()

        return updated

    # Correct some invalid identifiers spotted in the wild:
    # 1999/20 -> 1999/2000
    # 2000/2001 -> 2000/01
    # 1999/98 -> 1999/2000
    def sanitize_basefile(self, basefile):
        (y1, y2, idx) = re.split("[:/]", basefile)
        assert len(
            y1) == 4, "Basefile %s is invalid beyond sanitization" % basefile
        if y1 == "1999" and y2 != "2000":
            sanitized = "1999/2000:" + idx
            self.log.warning("Basefile given as %s, correcting to %s" %
                             (basefile, sanitized))
        elif (y1 != "1999" and
              (len(y2) != 2 or  # eg "2000/001"
               int(y1[2:]) + 1 != int(y2))):  # eg "1999/98

            sanitized = "%s/%02d:%s" % (y1, int(y1[2:]) + 1, idx)
            self.log.warning("Basefile given as %s, correcting to %s" %
                             (basefile, sanitized))
        else:
            sanitized = basefile
        return sanitized

    # For parsing:
    # 1999/94 and 1994/95 has only plaintext
    # 1995/96 to 2006/07 has plaintext + doc
    # 2007/08 onwards has plaintext, doc and pdf
    @managedparsing
    def parse(self, doc):
        try:

            # prefer PDF or Word files over the plaintext-containing HTML files
            # FIXME: PDF or Word files are now stored as attachments
            htmlfile = self.store.downloaded_path(doc.basefile)
            
            pdffile = self.store.path(doc.basefile, 'downloaded', '.pdf')
            
            wordfiles = (self.store.path(doc.basefile, 'downloaded', '.doc'),
                         self.store.path(doc.basefile, 'downloaded', '.docx'),
                         self.store.path(doc.basefile, 'downloaded', '.wpd'),
                         self.store.path(doc.basefile, 'downloaded', '.rtf'))

            # check if ANY of these exist
            if not filter(None, [os.path.exists(f) for f in wordfiles + (htmlfile, pdffile)]):
                raise errors.NoDownloadedFileError("File '%s' (or any .pdf/.doc/.docx/.wpd/.rdf variant) not found" % htmlfile)
            
            wordfile = None
            for f in wordfiles:
                if os.path.exists(f):
                    wordfile = f
            
            doc.uri = self.canonical_uri(doc.basefile)
            d = Describer(doc.meta, doc.uri)
            d.rdftype(self.rdf_type)

            # if we lack a .pdf file, use Open/LibreOffice to convert any
            # .wpd or .doc file to .pdf first
            if (wordfile
                    and not os.path.exists(pdffile)):
                intermediate_pdf = self.store.path(
                    doc.basefile, "intermediate", ".pdf")
                if not os.path.exists(intermediate_pdf):
                    cmdline = "%s --headless -convert-to pdf -outdir '%s' %s" % (self.config.get('soffice', 'soffice'),
                                                                                 os.path.dirname(
                                                                                     intermediate_pdf),
                                                                                 wordfile)
                    self.log.debug(
                        "%s: Converting to PDF: %s" % (doc.basefile, cmdline))
                    (ret, stdout, stderr) = util.runcmd(
                        cmdline, require_success=True)
                pdffile = intermediate_pdf

            if os.path.exists(pdffile):
                self.log.debug("%s: Using %s" % (doc.basefile, pdffile))
                intermediate_dir = os.path.dirname(
                    self.store.path(doc.basefile, 'intermediate', '.foo'))
                pdfreader = PDFReader()
                keep_xml = "bz2" if self.config.compress == "bz2" else True
                pdfreader.read(pdffile, intermediate_dir, keep_xml=keep_xml)
                self.parse_from_pdfreader(pdfreader, doc)
            else:
                downloaded_path = self.store.downloaded_path(doc.basefile)
                intermediate_path = self.store.path(
                    doc.basefile, 'intermediate', '.txt')
                self.log.debug("%s: Using %s (%s)" % (doc.basefile,
                               downloaded_path, intermediate_path))
                if not os.path.exists(intermediate_path):
                    html = codecs.open(
                        downloaded_path, encoding="iso-8859-1").read()
                    util.writefile(intermediate_path, util.extract_text(
                        html, '<pre>', '</pre>'), encoding="utf-8")
                textreader = TextReader(intermediate_path, encoding="utf-8")
                self.parse_from_textreader(textreader, doc)
                # How to represent that one XHTML doc was created from
                # plaintext, and another from PDF? create a bnode
                # representing the source prov:wasDerivedFrom and set its
                # dct:format to correct mime type

            d.value(self.ns['prov'].wasGeneratedBy, self.qualified_class_name())
            self.infer_triples(d, doc.basefile)
            return True
        except Exception as e:
            err = errors.ParseError(str(e))
            if isinstance(e, IOError):
                err.dummyfile = self.store.parsed_path(doc.basefile)
            raise err

    def parse_from_pdfreader(self, pdfreader, doc):
        parser = offtryck_parser(preset='proposition')
        doc.body = parser.parse(pdfreader.textboxes(offtryck_gluefunc))
        

            
    def parse_from_textreader(self, textreader, doc):
        describer = Describer(doc.meta, doc.uri)
        for p in textreader.getiterator(textreader.readparagraph):
            # print "Handing %r (%s)" % (p[:40], len(doc.body))
            if not p.strip():
                continue
            elif not doc.body and 'Obs! Dokumenten i denna databas kan vara ofullständiga.' in p:
                continue
            elif not doc.body and p.strip().startswith("Dokument:"):
                # We already know this
                continue
            elif not doc.body and p.strip().startswith("Titel:"):
                describer.value(
                    self.ns['dct'].title, util.normalize_space(p[7:]))
            else:
                doc.body.append(Preformatted([p]))


class PropRiksdagen(Riksdagen):
    alias = "propriksdagen"
    rdf_type = RPUBL.Proposition
    document_type = Riksdagen.PROPOSITION

# inherit list_basefiles_for from CompositeStore, basefile_to_pathfrag
# from SwedishLegalStore)
class PropositionerStore(CompositeStore, SwedishLegalStore):
    pass

class Propositioner(CompositeRepository, SwedishLegalSource):
    subrepos = PropPolo, PropTrips, PropRiksdagen
    alias = "prop"
    xslt_template = "paged.xsl"
    storage_policy = "dir"
    rdf_type = RPUBL.Proposition
    documentstore_class = PropositionerStore
    def tabs(self, primary=False):
        return [('Förarbeten', self.dataset_uri())]
